"""
Test the uuid queue module adapter queue file
- _get_combined_uuids_gen
- _get_uncombined_uuids
- QueueTypes
- QueueAdapter
- WorkerAdapter
"""
import uuid

from unittest import (
    TestCase,
    mock,
)

import pytest

from ..adapter_queue import (
    _get_combined_uuids_gen,
    _get_uncombined_uuids,
    QueueTypes,
    QueueAdapter,
    WorkerAdapter,
)

from ..queues.base_queue import (
    BASE_QUEUE_TYPE,
    BaseQueueClient,
    BaseQueue,
)
from ..queues.redis_queues import (
    RedisClient,
)
from .test_redis_queue import (
    REDIS_QUEUES,
    REDIS_QUEUE_CLASSES,
)

MOCK_TIME = 123456.654321
REDIS_QUEUE_TYPES = [a for a, b in REDIS_QUEUES]
UUID_LENGTH = len('75659c39-eaea-47b7-ba26-92e9ff183e6c')


# Test Fixtures
# Test Helpers
def _get_uuids(cnt):
    """
    Return things like this 75659c39-eaea-47b7-ba26-92e9ff183e6c
    """
    uuids = set()
    while len(uuids) < cnt:
        uuids.add(str(uuid.uuid4()))
    return uuids


# _get_combined_uuids_gen
def test_getcombids_wrong_len():
    """Test _get_combined_uuids_gen with different uuid lengths"""
    uuids_count = 17
    uuids_set = _get_uuids(uuids_count)
    batch_by = 17
    uuid_len = UUID_LENGTH
    max_batch_size = batch_by * uuids_count * uuid_len
    uuids_list = []
    for uuid_str in uuids_set:
        uuids_list.append(uuid_str[:uuid_len - 1])
    with pytest.raises(ValueError):
        items = _get_combined_uuids_gen(
            batch_by,
            uuid_len,
            max_batch_size,
            uuids_list
        )
        # Must run generator
        for item in items:  # pylint: disable=unused-variable
            pass


def test_getcombids_one_batch():
    """Test _get_combined_uuids_gen one batch"""
    uuids_count = 17
    uuids_set = _get_uuids(uuids_count)
    batch_by = uuids_count
    uuid_len = UUID_LENGTH
    max_batch_size = batch_by * uuid_len
    combined_uuids = _get_combined_uuids_gen(
        batch_by,
        uuid_len,
        max_batch_size,
        uuids_set
    )
    expected_combined_uuids_count = 1
    expected_full_length = 1
    expected_partial_length = 0
    combined_uuids_count = 0
    full_length = 0
    partial_length = 0
    for comb_uuid_str in combined_uuids:
        combined_uuids_count += 1
        if len(comb_uuid_str) == uuid_len * batch_by:
            full_length += 1
        else:
            partial_length += 1
    assert combined_uuids_count == expected_combined_uuids_count
    assert expected_full_length == full_length
    assert expected_partial_length == partial_length


def test_getcombids_two_batch_max():
    """Test _get_combined_uuids_gen two batch via max size"""
    uuids_count = 17
    uuids_set = _get_uuids(uuids_count)
    batch_by = 17
    uuid_len = UUID_LENGTH
    max_batch_size = batch_by * uuid_len - 1
    combined_uuids = _get_combined_uuids_gen(
        batch_by,
        uuid_len,
        max_batch_size,
        uuids_set
    )
    expected_combined_uuids_count = 2
    expected_full_length = 1
    expected_partial_length = 1
    combined_uuids_count = 0
    full_length = 0
    partial_length = 0
    for comb_uuid_str in combined_uuids:
        combined_uuids_count += 1
        if len(comb_uuid_str) == uuid_len * (batch_by - 1):
            full_length += 1
        else:
            partial_length += 1
    assert combined_uuids_count == expected_combined_uuids_count
    assert expected_full_length == full_length
    assert expected_partial_length == partial_length


def test_getcombids_many_batch():
    """Test _get_combined_uuids_gen with many batches"""
    uuids_count = 47
    uuids_set = _get_uuids(uuids_count)
    batch_by = 6
    uuid_len = UUID_LENGTH
    max_batch_size = batch_by * uuid_len
    combined_uuids = _get_combined_uuids_gen(
        batch_by,
        uuid_len,
        max_batch_size,
        uuids_set
    )
    expected_combined_uuids_count = uuids_count//batch_by
    if uuids_count % batch_by > 0:
        expected_combined_uuids_count += 1
    expected_full_length = expected_combined_uuids_count - 1
    expected_partial_length = 1
    combined_uuids_count = 0
    full_length = 0
    partial_length = 0
    for comb_uuid_str in combined_uuids:
        combined_uuids_count += 1
        if len(comb_uuid_str) == uuid_len * batch_by:
            full_length += 1
        else:
            partial_length += 1
    assert combined_uuids_count == expected_combined_uuids_count
    assert expected_full_length == full_length
    assert expected_partial_length == partial_length


# _get_uncombined_uuids
def test_get_uncombined_uuids():
    """Test _get_uncombined_uuids"""
    uuids_count = 47
    uuids_list = list(_get_uuids(uuids_count))
    batch_by = 6
    uuid_len = UUID_LENGTH
    max_batch_size = batch_by * uuid_len
    combined_uuids = _get_combined_uuids_gen(
        batch_by,
        uuid_len,
        max_batch_size,
        uuids_list
    )
    uncombined_uuids = []
    for combined_uuid in combined_uuids:
        uncombined_uuids.extend(_get_uncombined_uuids(uuid_len, combined_uuid))
    uuids_list.sort()
    uncombined_uuids.sort()
    assert uuids_list == uncombined_uuids


class TestQueueTypes(TestCase):
    """
    Test QueueTypes class

    Everything is a classmethod or class var
    """
    @classmethod
    def setUpClass(cls):
        cls.queue_types = [
            queue_set[0]
            for queue_set in QueueTypes.QUEUES
        ]
        cls.queue_types.sort()

    def test_init(self):
        """
        Test available queue types and clients
        - This should failed when new queue types are added
        """
        for queue_set in QueueTypes.QUEUES:
            queue_type, queue_client = queue_set
            if queue_type == BASE_QUEUE_TYPE:
                self.assertTrue(queue_client is BaseQueueClient)
            elif queue_type in REDIS_QUEUE_TYPES:
                self.assertTrue(queue_client is RedisClient)
            else:
                print("New Queue(%s) not handled in test" % queue_type)
                assert False

    def test_check_queue_type_true(self):
        """Test check_queue_type with real queue type"""
        for queue_type in self.queue_types:
            self.assertTrue(QueueTypes.check_queue_type(queue_type))

    def test_check_queue_type_false(self):
        """Test check_queue_type with fake queue type"""
        queue_type = 'not a queue type'
        self.assertFalse(QueueTypes.check_queue_type(queue_type))

    def test_get_all(self):
        """Test get_all"""
        queue_types = QueueTypes.get_all()
        queue_types.sort()
        self.assertListEqual(queue_types, self.queue_types)

    def test_get_queue_client_class(self):
        """Test get_queue_client_class"""
        for queue_set in QueueTypes.QUEUES:
            queue_type, expected_queue_client = queue_set
            queue_client = QueueTypes.get_queue_client_class(queue_type)
            self.assertTrue(expected_queue_client is queue_client)

    def test_redis_queue_types(self):
        """
        Test that redis queue types exist
        """
        for queue_type in REDIS_QUEUE_TYPES:
            self.assertTrue(
                QueueTypes.check_queue_type(queue_type)
            )


class TestQueueAdapter(TestCase):
    """Test QueueAdapter - Uuid Server"""
    # pylint: disable=too-many-public-methods
    members = [
        ('_queue_name', str),
        ('_queue_type', str),
        ('_queue_options', dict),
        ('_start_us', int),
        ('queue_id', str),
        ('_queue', None), # depends on _queue_type
    ]
    meth_func = [
        # Errors
        '_has_errors',
        'pop_errors',
        # Worker
        '_get_worker_id',
        '_get_worker_queue',
        'get_worker',
        'update_worker_conn',
        # Uuids
        'has_uuids',
        '_load_uuids',
        'load_uuids',
        # Run
        'is_indexing',
        # Queue Client
        '_get_queue',
        'close_indexing', # not tested
    ]

    @classmethod
    @mock.patch('time.time', mock.MagicMock(return_value=MOCK_TIME))
    def setUpClass(cls):
        '''Create a simple queue to be used most tests'''
        cls.load_uuids = [str(num) for num in range(10, 99)]
        cls.queue_name = 'test-queue-name'
        cls.queue_type = BASE_QUEUE_TYPE
        cls.queue_options = {
            'processes': 1,
            'chunk_size': 1,
            'batch_size': 10,
            'uuid_len': 2,
        }
        cls.mock_time_us = int(MOCK_TIME * 1000000)
        cls.queue_server = QueueAdapter(
            cls.queue_name,
            cls.queue_type,
            cls.queue_options
        )
        cls.queue_worker = None

    def setUp(self):
        """Setup for each test"""
        self.queue_worker = self.queue_server.get_worker()

    def tearDown(self):
        """Reset for each test"""
        self.queue_worker = None
        # pylint: disable=protected-access
        self.queue_server._queue._qmeta._errors = []
        self.queue_server._queue._qmeta._worker_conns = {}
        self.queue_server._queue._qmeta._worker_results = {}
        self.queue_server._queue._qmeta._uuid_count = 0
        self.queue_server._queue._uuids = []

    # private/protected
    def test_init_dir(self):
        '''Test SimpleUuidServer has expected function and variables'''
        dir_items = [
            item
            for item in dir(self.queue_server)
            if item[0:2] != '__'
        ]
        member_names = [name for name, _ in self.members]
        for item in dir_items:
            if item in member_names:
                continue
            if item in self.meth_func:
                continue
            raise AssertionError('Member or method not being tested:%s' % item)
        for item, item_type in self.members:
            self.assertTrue(hasattr(self.queue_server, item))
            if item_type:
                self.assertIsInstance(
                    getattr(self.queue_server, item),
                    item_type
                )
        for item in self.meth_func:
            self.assertTrue(hasattr(self.queue_server, item))
            self.assertTrue(
                str(type(getattr(self.queue_server, item))) in [
                    "<class 'method'>",
                    "<class 'function'>"
                ]
            )

    def test_init_basic(self):
        '''Test _get_worker_id'''
        self.assertEqual(str(self.mock_time_us), self.queue_server.queue_id)
        # pylint: disable=protected-access
        self.assertEqual(self.mock_time_us, self.queue_server._start_us)
        self.assertDictEqual(
            self.queue_options,
            self.queue_server._queue_options
        )

    # Errors
    def test_has_errors_false(self):
        '''Test _has_errors returns false when empty'''
        # pylint: disable=protected-access
        self.assertFalse(self.queue_server._has_errors())

    def test_has_errors_true(self):
        '''Test _has_errors return true when not empty'''
        # pylint: disable=protected-access
        self.queue_server._queue._qmeta._errors = ['e']
        self.assertTrue(self.queue_server._has_errors())

    def test_pop_errors(self):
        '''Test pop_errors'''
        given_errors = ['e1', 'e2', 'e3']
        # pylint: disable=protected-access
        self.queue_server._queue._qmeta._errors = given_errors
        errors = self.queue_server.pop_errors()
        given_errors.sort()
        self.assertListEqual(errors, given_errors)
        self.assertListEqual(
            [],
            self.queue_server._queue._qmeta._errors
        )

    def test_get_worker_id(self):
        '''Test _get_worker_id'''
        # Plus two since a worker is added in setUp
        id_time_us = self.mock_time_us + 2
        # pylint: disable=protected-access
        worker_id = self.queue_server._get_worker_id()
        self.assertTrue(
            worker_id == str(id_time_us)
        )
        _ = self.queue_server._get_worker_id()
        # The function does not add the worker to the conn
        # so the id should not change
        worker_id = self.queue_server._get_worker_id()
        self.assertTrue(
            worker_id == str(id_time_us)
        )

    def test_get_worker_queue_base(self):
        '''Test _get_worker_queue for base queue type'''
        queue_name = 'test-wrk-queue-name'
        queue_type = BASE_QUEUE_TYPE
        queue_options = {
            'processes': 1,
            'chunk_size': 1,
            'batch_size': 10,
            'uuid_len': 2,
        }
        tmp_queue_server = QueueAdapter(
            queue_name,
            queue_type,
            queue_options
        )
        # pylint: disable=protected-access
        tmp_worker_queue = tmp_queue_server._get_worker_queue()
        self.assertTrue(
            tmp_queue_server._queue is tmp_worker_queue
        )

    def test_get_worker_queue_redis(self):
        '''Test _get_worker_queue for redis queue types'''
        for queue_type, queue_class in REDIS_QUEUE_CLASSES:
            queue_name = queue_type.lower()
            queue_type = queue_type
            queue_options = {
                'processes': 1,
                'chunk_size': 1,
                'batch_size': 10,
                'uuid_len': 2,
                'host': 'localhost',
                'port': 6379,
            }
            tmp_queue_server = QueueAdapter(
                queue_name,
                queue_type,
                queue_options
            )
            # pylint: disable=protected-access
            self.assertIsInstance(tmp_queue_server._queue, queue_class)

    def test_get_worker_base(self):
        '''Test get_worker with base queue'''
        queue_name = 'test-wrk-queue-name'
        queue_type = BASE_QUEUE_TYPE
        queue_options = {
            'processes': 1,
            'chunk_size': 1,
            'batch_size': 10,
            'uuid_len': 2,
        }
        tmp_queue_server = QueueAdapter(
            queue_name,
            queue_type,
            queue_options
        )
        tmp_worker_queue = tmp_queue_server.get_worker()
        self.assertIsInstance(tmp_worker_queue, WorkerAdapter)
        # pylint: disable=protected-access
        ref_worker_id = str(tmp_queue_server._start_us + 1)
        self.assertEqual(ref_worker_id, tmp_worker_queue.worker_id)
        self.assertTrue(
            ref_worker_id in tmp_queue_server._queue._qmeta._worker_conns
        )

    def test_get_worker_redis(self):
        '''Test get_worker for redis queues'''
        for queue_type, queue_class in REDIS_QUEUE_CLASSES:
            queue_name = queue_type.lower()
            queue_type = queue_type
            queue_options = {
                'processes': 1,
                'chunk_size': 1,
                'batch_size': 10,
                'uuid_len': 2,
                'host': 'localhost',
                'port': 6379,
            }
            tmp_queue_server = QueueAdapter(
                queue_name,
                queue_type,
                queue_options
            )
            # pylint: disable=protected-access
            self.assertIsInstance(tmp_queue_server._queue, queue_class)

    def test_update_worker_conn(self):
        '''Test update_worker_conn'''
        update_uuid_cnt = 9
        update_get_cnt = 3
        worker_id = self.queue_worker.worker_id
        # pylint: disable=protected-access
        worker_conn = self.queue_server._queue._qmeta._worker_conns[worker_id]
        self.queue_server.update_worker_conn(
            worker_id,
            update_uuid_cnt,
            update_get_cnt,
        )
        self.assertDictEqual(
            worker_conn,
            {
                'uuid_cnt': update_uuid_cnt,
                'get_cnt': update_get_cnt
            }
        )

    # Uuids
    def test_has_uuids(self):
        '''Test has_uuids'''
        # pylint: disable=protected-access
        self.queue_server._queue._qmeta._uuid_count = 4
        self.assertTrue(self.queue_server.has_uuids())

    def test_has_uuids_empty(self):
        '''Test has_uuids when empty'''
        self.assertFalse(self.queue_server.has_uuids())

    def test_load_uuids_batch_one(self):
        '''Test _load_uuids with batch size 1'''
        queue_name = 'test-wrk-queue-name'
        queue_type = BASE_QUEUE_TYPE
        queue_options = {
            'processes': 1,
            'chunk_size': 1,
            'batch_size': 1,
            'uuid_len': 2,
        }
        tmp_queue_server = QueueAdapter(
            queue_name,
            queue_type,
            queue_options
        )
        load_uuids_len = len(self.load_uuids)
        # pylint: disable=protected-access
        success_cnt = tmp_queue_server._load_uuids(self.load_uuids.copy())
        self.assertEqual(load_uuids_len, success_cnt)
        self.assertEqual(
            len(tmp_queue_server._queue._uuids),
            load_uuids_len
        )

    def test_load_uuids_batch_many(self):
        '''Test _load_uuids with batch size greater than 1'''
        load_uuids_len = len(self.load_uuids.copy())
        batch_size = self.queue_options['batch_size']
        expected_batches = load_uuids_len//batch_size
        if load_uuids_len % batch_size:
            expected_batches += 1
        # pylint: disable=protected-access
        success_cnt = self.queue_server._load_uuids(self.load_uuids.copy())
        self.assertEqual(load_uuids_len, success_cnt)
        self.assertEqual(
            len(self.queue_server._queue._uuids),
            expected_batches
        )

    def test_load_uuids_indexing(self):
        '''Test load_uuids when already indexing'''
        # pylint: disable=protected-access
        self.queue_server._queue._qmeta._uuid_count = 583
        self.assertIsNone(self.queue_server.load_uuids(self.load_uuids.copy()))

    def test_load_uuids(self):
        '''Test load_uuids'''
        copy_load_uuids = self.load_uuids.copy()
        load_uuid_len = len(self.load_uuids)
        _ = self.queue_server.load_uuids(copy_load_uuids)
        # pylint: disable=protected-access
        self.assertEqual(
            self.queue_server._queue._qmeta._uuid_count,
            load_uuid_len
        )

    # Run
    def test_is_indexing_false_all(self):
        '''
        Test is_indexing is false
        -No uuids, no errors, and worker uuid cnt is 0
        '''
        self.assertFalse(self.queue_server.is_indexing())

    def test_is_indexing_true_uuidcnt(self):
        '''
        Test is_indexing is true when work uuid cnt is not 0
        '''
        worker_id = self.queue_worker.worker_id
        # pylint: disable=protected-access
        self.queue_server._queue._qmeta._worker_conns[worker_id]['uuid_cnt'] = 1
        self.assertTrue(self.queue_server.is_indexing())

    def test_is_indexing_true_haserrors(self):
        '''
        Test is_indexing is true when has errors is true
        '''
        # pylint: disable=protected-access
        self.queue_server._queue._qmeta._errors = [1]
        self.assertTrue(self.queue_server.is_indexing())

    def test_is_indexing_true_hasuuids(self):
        '''
        Test is_indexing is true when has uuids is true
        '''
        # pylint: disable=protected-access
        self.queue_server._queue._qmeta._uuid_count = 130
        self.assertTrue(self.queue_server.is_indexing())

    # Queue Client
    def test_get_queue_base(self):
        '''Test _get_queue for base queue'''
        # pylint: disable=protected-access
        self.assertIsInstance(self.queue_server._get_queue(self.queue_name), BaseQueue)

    def test_get_queue_redis(self):
        '''Test _get_queue for redis queues'''
        for queue_type, queue_class in REDIS_QUEUE_CLASSES:
            queue_name = 'test-wrk-queue-name'
            queue_type = queue_type
            queue_options = {
                'processes': 1,
                'chunk_size': 1,
                'batch_size': 1,
                'uuid_len': 2,
                'host': 'localhost',
                'port': 6379,
            }
            tmp_queue_server = QueueAdapter(
                queue_name,
                queue_type,
                queue_options
            )
            # pylint: disable=protected-access
            self.assertIsInstance(tmp_queue_server._queue, queue_class)
            self.assertIsInstance(tmp_queue_server._get_queue(queue_name), queue_class)

    def test_get_queue_bad_type(self):
        '''Test _get_queue'''
        queue_name = 'test-wrk-queue-name'
        queue_type = 'bad-queue-type'
        queue_options = {
            'processes': 1,
            'chunk_size': 1,
            'batch_size': 1,
            'uuid_len': 2,
        }
        tmp_queue_server = QueueAdapter(
            queue_name,
            queue_type,
            queue_options
        )
        # pylint: disable=protected-access
        self.assertIsNone(tmp_queue_server._queue)
        self.assertIsNone(tmp_queue_server._get_queue(queue_name))


class TestWorkerAdapater(TestCase):
    """Test WorkerAdapater - Uuid Worker"""
    # pylint: disable=too-many-public-methods
    members = [
        ('_queue_name', str),
        ('_queue_options', dict),
        ('worker_id', str),
        ('_queue', None), # depends on _queue_type
        ('is_running', bool),
        ('get_cnt', int),
        ('uuid_cnt', int),
        ('chunk_size', int),
        ('processes', int),
    ]
    meth_func = [
        # Uuids
        '_get_uncombined_uuids',
        '_get_uuids',
        'get_uuids',
        # Run
        'update_finished',
    ]

    @staticmethod
    def _get_worker(*args):
        (queue_name, queue_type,
         queue_options, worker_id) = args
        client_class = QueueTypes.get_queue_client_class(queue_type)
        client = client_class(queue_options)
        queue = client.get_queue(queue_name, queue_type)
        return WorkerAdapter(
            queue_name,
            queue_options,
            worker_id,
            queue,
        )

    @classmethod
    @mock.patch('time.time', mock.MagicMock(return_value=MOCK_TIME))
    def setUpClass(cls):
        '''Create a simple worker to be used most tests'''
        cls.base_queue_name = 'base-queue-name'
        cls.base_queue_options = {
            'processes': 1,
            'chunk_size': 1,
            'batch_size': 10,
            'uuid_len': 4,
            'get_size': 1,
        }
        cls.base_worker_id = 'base-worker-id'
        cls.base_worker = cls._get_worker(
            cls.base_queue_name,
            BASE_QUEUE_TYPE,
            cls.base_queue_options,
            cls.base_worker_id,
        )

    def setUp(self):
        """Setup for each test"""
        pass

    def tearDown(self):
        """Reset for each test"""
        pass

    # private/protected
    def test_init_dir(self):
        '''Test SimpleUuidServer has expected function and variables'''
        dir_items = [
            item
            for item in dir(self.base_worker)
            if item[0:2] != '__'
        ]
        member_names = [name for name, _ in self.members]
        for item in dir_items:
            if item in member_names:
                continue
            if item in self.meth_func:
                continue
            raise AssertionError('Member or method not being tested:%s' % item)
        for item, item_type in self.members:
            self.assertTrue(hasattr(self.base_worker, item))
            if item_type:
                self.assertIsInstance(
                    getattr(self.base_worker, item),
                    item_type
                )
        for item in self.meth_func:
            self.assertTrue(hasattr(self.base_worker, item))
            self.assertTrue(
                str(type(getattr(self.base_worker, item))) in [
                    "<class 'method'>",
                    "<class 'function'>",
                    "<class 'unittest.mock.MagicMock'>",
                ]
            )

    def test_init_basic(self):
        '''Test _get_worker_id'''
        # pylint: disable=protected-access
        self.assertEqual(self.base_queue_name, self.base_worker._queue_name)
        self.assertDictEqual(
            self.base_queue_options,
            self.base_worker._queue_options
        )
        self.assertEqual(self.base_worker_id, self.base_worker.worker_id)
        self.assertIsInstance(self.base_worker._queue, BaseQueue)
        self.assertFalse(self.base_worker.is_running)
        self.assertEqual(0, self.base_worker.get_cnt)
        self.assertEqual(0, self.base_worker.uuid_cnt)


    def test_get_uuids_batchsize_one(self):
        '''Test _get_uuids when batch size is one'''
        # pylint: disable=protected-access
        self.base_worker._queue_options['batch_size'] = 1
        org_queue_get_uuids = self.base_worker._queue.get_uuids
        org_get_uncombined_uuids = self.base_worker._get_uncombined_uuids
        self.base_worker._queue.get_uuids = mock.MagicMock()
        self.base_worker._get_uncombined_uuids = mock.MagicMock()
        _ = self.base_worker._get_uuids()
        self.assertEqual(1, self.base_worker._queue.get_uuids.call_count)
        self.base_worker._queue.get_uuids.assert_called_with(1)
        self.assertEqual(0, self.base_worker._get_uncombined_uuids.call_count)
        self.base_worker._queue.get_uuids = org_queue_get_uuids
        self.base_worker._get_uncombined_uuids = org_get_uncombined_uuids


    def test_get_uuids_batchsize_many(self):
        '''Test _get_uuids when batch size is greater than one'''
        # pylint: disable=protected-access
        self.base_worker._queue_options['batch_size'] = 2
        return_uuids = ['u01xu02x']
        expected_result = ['u01x', 'u02x']
        org_queue_get_uuids = self.base_worker._queue.get_uuids
        org_get_uncombined_uuids = self.base_worker._get_uncombined_uuids
        self.base_worker._queue.get_uuids = mock.MagicMock(
            return_value=return_uuids
        )
        self.base_worker._get_uncombined_uuids = mock.MagicMock(
            return_value=expected_result
        )
        result = self.base_worker._get_uuids()
        self.assertEqual(1, self.base_worker._queue.get_uuids.call_count)
        self.base_worker._queue.get_uuids.assert_called_with(1)
        self.assertEqual(1, self.base_worker._get_uncombined_uuids.call_count)
        self.base_worker._get_uncombined_uuids.assert_called_with(
            4,
            return_uuids[0]
        )
        result.sort()
        expected_result.sort()
        self.assertListEqual(result, expected_result)
        self.base_worker._queue.get_uuids = org_queue_get_uuids
        self.base_worker._get_uncombined_uuids = org_get_uncombined_uuids

    def test_get_uuids_uuid_cnt_zero(self):
        '''Test get_uuids when uuid cnt is zero'''
        get_return_value = ['u01x', 'u02x']
        queue_options = self.base_queue_options.copy()
        base_worker = self._get_worker(
            self.base_queue_name,
            BASE_QUEUE_TYPE,
            queue_options,
            self.base_worker_id,
        )
        # pylint: disable=protected-access
        base_worker._get_uuids = mock.MagicMock(
            return_value=get_return_value
        )
        base_worker._queue.update_uuid_count = mock.MagicMock()
        base_worker._queue.update_worker_conn = mock.MagicMock()
        base_worker.get_uuids()
        self.assertEqual(base_worker.uuid_cnt, len(get_return_value))
        self.assertEqual(1, base_worker._queue.update_uuid_count.call_count)
        base_worker._queue.update_uuid_count.assert_called_with(-1 * len(get_return_value))
        self.assertEqual(1, base_worker._queue.update_worker_conn.call_count)
        base_worker._queue.update_worker_conn.assert_called_with(
            base_worker.worker_id,
            len(get_return_value),
            1,
        )

    def test_get_uuids_uuid_cnt(self):
        '''Test get_uuids when uuid cnt is not zero'''
        queue_options = self.base_queue_options.copy()
        base_worker = self._get_worker(
            self.base_queue_name,
            BASE_QUEUE_TYPE,
            queue_options,
            self.base_worker_id,
        )
        base_worker.get_cnt = 1
        base_worker.uuid_cnt = 14
        # pylint: disable=protected-access
        base_worker._get_uuids = mock.MagicMock(return_value=[])
        base_worker._queue.update_uuid_count = mock.MagicMock()
        base_worker._queue.update_worker_conn = mock.MagicMock()
        base_worker.get_uuids()
        self.assertEqual(base_worker.uuid_cnt, 14)
        self.assertEqual(0, base_worker._queue.update_uuid_count.call_count)
        self.assertEqual(1, base_worker._queue.update_worker_conn.call_count)
        base_worker._queue.update_worker_conn.assert_called_with(
            base_worker.worker_id,
            14,
            2,
        )

    def test_update_finished_okay(self):
        '''Test update_finished when Okay finish'''
        base_worker = self._get_worker(
            self.base_queue_name,
            BASE_QUEUE_TYPE,
            self.base_queue_options,
            self.base_worker_id,
        )
        base_worker.uuid_cnt = 4
        return_msg = 'Okay'
        # pylint: disable=protected-access
        base_worker._queue.update_finished = mock.MagicMock(
            return_value=return_msg
        )
        exepcted_result_msg = None  # None when 'Okay'
        given_results = {'test': 1, 'successes': 1, 'errors': []}
        result_msg = base_worker.update_finished(given_results)
        self.assertEqual(1, base_worker._queue.update_finished.call_count)
        base_worker._queue.update_finished.assert_called_with(
            base_worker.worker_id,
            given_results,
        )
        self.assertEqual(exepcted_result_msg, result_msg)
        self.assertEqual(0, base_worker.uuid_cnt)

    def test_update_finished_not_okay(self):
        '''Test update_finished when not Okay finish'''
        base_worker = self._get_worker(
            self.base_queue_name,
            BASE_QUEUE_TYPE,
            self.base_queue_options,
            self.base_worker_id,
        )
        base_worker.uuid_cnt = 4
        return_msg = 'Not Okay'
        # pylint: disable=protected-access
        base_worker._queue.update_finished = mock.MagicMock(
            return_value=return_msg
        )
        exepcted_result_msg = 'Update finished could not reset worker: %s' % return_msg
        given_results = {'test': 1, 'successes': 1, 'errors': []}
        result_msg = base_worker.update_finished(given_results)
        self.assertEqual(1, base_worker._queue.update_finished.call_count)
        base_worker._queue.update_finished.assert_called_with(
            base_worker.worker_id,
            given_results,
        )
        self.assertEqual(exepcted_result_msg, result_msg)
        self.assertEqual(4, base_worker.uuid_cnt)



from functools import lru_cache
from pyramid.security import (
    ALL_PERMISSIONS,
    Allow,
    Authenticated,
    Deny,
    DENY_ALL,
    Everyone,
)
from pyramid.traversal import (
    find_root,
    traverse,
)
import snovault
from ..schema_formats import is_accession
from snovault.elasticsearch.searches.configs import search_config


@lru_cache()
def _award_viewing_group(award_uuid, root):
    award = root.get_by_uuid(award_uuid)
    return award.upgrade_properties().get('viewing_group')

# Item acls

ONLY_ADMIN_VIEW = [
    (Allow, 'group.admin', ['view', 'edit']),
    (Allow, 'group.read-only-admin', ['view']),
    (Allow, 'remoteuser.INDEXER', ['view']),
    (Allow, 'remoteuser.EMBED', ['view']),
    (Deny, Everyone, ['view', 'edit']),
]

ALLOW_EVERYONE_VIEW = [
    (Allow, Everyone, 'view'),
] + ONLY_ADMIN_VIEW


ALLOW_VIEWING_GROUP_VIEW = [
    (Allow, 'role.viewing_group_member', 'view'),
] + ONLY_ADMIN_VIEW

ALLOW_LAB_SUBMITTER_EDIT = [
    (Allow, 'role.viewing_group_member', 'view'),
    (Allow, 'role.lab_submitter', 'edit'),
] + ONLY_ADMIN_VIEW

ALLOW_CURRENT_AND_SUBMITTER_EDIT = [
    (Allow, Everyone, 'view'),
    (Allow, 'role.lab_submitter', 'edit'),
] + ONLY_ADMIN_VIEW

ALLOW_CURRENT = [
    (Allow, Everyone, 'view'),
] + ONLY_ADMIN_VIEW

DELETED = [
    (Deny, Everyone, 'visible_for_edit')
] + ONLY_ADMIN_VIEW


# Collection acls

ALLOW_SUBMITTER_ADD = [
    (Allow, 'group.submitter', ['add']),
]



def paths_filtered_by_status(request, paths, exclude=('deleted', 'replaced'), include=None):
    if include is not None:
        return [
            path for path in paths
            if traverse(request.root, path)['context'].__json__(request).get('status') in include
        ]
    else:
        return [
            path for path in paths
            if traverse(request.root, path)['context'].__json__(request).get('status') not in exclude
        ]


class AbstractCollection(snovault.AbstractCollection):
    def get(self, name, default=None):
        resource = super(AbstractCollection, self).get(name, None)
        if resource is not None:
            return resource
        if ':' in name:
            resource = self.connection.get_by_unique_key('alias', name)
            if resource is not None:
                if not self._allow_contained(resource):
                    return default
                return resource
        return default


class Collection(snovault.Collection, AbstractCollection):
    def __init__(self, *args, **kw):
        super(Collection, self).__init__(*args, **kw)
        if hasattr(self, '__acl__'):
            return
        # XXX collections should be setup after all types are registered.
        # Don't access type_info.schema here as that precaches calculated schema too early.
        if 'lab' in self.type_info.factory.schema['properties']:
            self.__acl__ = ALLOW_SUBMITTER_ADD


class Item(snovault.Item):
    AbstractCollection = AbstractCollection
    Collection = Collection
    STATUS_ACL = {
        # standard_status
        'released': ALLOW_CURRENT,
        'deleted': DELETED,
        'replaced': DELETED,

        # shared_status
        'current': ALLOW_CURRENT,
        'disabled': ONLY_ADMIN_VIEW,

        # file
        'obsolete': ONLY_ADMIN_VIEW,

        # "sets"
        'release ready': ALLOW_VIEWING_GROUP_VIEW,
        'revoked': ALLOW_CURRENT,
        'in review': ALLOW_CURRENT_AND_SUBMITTER_EDIT,

        # publication
        'published': ALLOW_CURRENT,

        # pipeline
        'active': ALLOW_CURRENT,
        'archived': ALLOW_CURRENT,
    }

    @property
    def __name__(self):
        if self.name_key is None:
            return self.uuid
        properties = self.upgrade_properties()
        if properties.get('status') == 'replaced':
            return self.uuid
        return properties.get(self.name_key, None) or self.uuid

    def __acl__(self):
        # Don't finalize to avoid validation here.
        properties = self.upgrade_properties().copy()
        status = properties.get('status')
        return self.STATUS_ACL.get(status, ALLOW_LAB_SUBMITTER_EDIT)

    def __ac_local_roles__(self):
        roles = {}
        properties = self.upgrade_properties().copy()
        if 'lab' in properties:
            lab_submitters = 'submits_for.%s' % properties['lab']
            roles[lab_submitters] = 'role.lab_submitter'
        if 'award' in properties:
            viewing_group = _award_viewing_group(properties['award'], find_root(self))
            if viewing_group is not None:
                viewing_group_members = 'viewing_group.%s' % viewing_group
                roles[viewing_group_members] = 'role.viewing_group_member'
        return roles

    def unique_keys(self, properties):
        keys = super(Item, self).unique_keys(properties)
        if 'accession' not in self.schema['properties']:
            return keys
        keys.setdefault('accession', []).extend(properties.get('alternate_accessions', []))
        if properties.get('status') != 'replaced' and 'accession' in properties:
            keys['accession'].append(properties['accession'])
        return keys


@search_config(
    name='Item'
)
def item_search_config():
    return {}


class SharedItem(Item):
    ''' An Item visible to all authenticated users while "proposed" or "in progress".
    '''
    def __ac_local_roles__(self):
        roles = {}
        properties = self.upgrade_properties().copy()
        if 'lab' in properties:
            lab_submitters = 'submits_for.%s' % properties['lab']
            roles[lab_submitters] = 'role.lab_submitter'
        roles[Authenticated] = 'role.viewing_group_member'
        return roles


@snovault.calculated_property(context=Item.Collection, category='action')
def add(context, request):
    if request.has_permission('add'):
        return {
            'name': 'add',
            'title': 'Add',
            'profile': '/profiles/{ti.name}.json'.format(ti=context.type_info),
            'href': '{item_uri}#!add'.format(item_uri=request.resource_path(context)),
        }


@snovault.calculated_property(context=Item, category='action')
def edit(context, request):
    if request.has_permission('edit'):
        return {
            'name': 'edit',
            'title': 'Edit',
            'profile': '/profiles/{ti.name}.json'.format(ti=context.type_info),
            'href': '{item_uri}#!edit'.format(item_uri=request.resource_path(context)),
        }


@snovault.calculated_property(context=Item, category='action')
def edit_json(context, request):
    if request.has_permission('edit'):
        return {
            'name': 'edit-json',
            'title': 'Edit JSON',
            'profile': '/profiles/{ti.name}.json'.format(ti=context.type_info),
            'href': '{item_uri}#!edit-json'.format(item_uri=request.resource_path(context)),
        }



def test_graph_dot(testapp):
    res = testapp.get('/profiles/graph.dot', status=200)
    assert res.content_type == 'text/vnd.graphviz'
    assert res.text


def test_graph_svg(testapp):
    res = testapp.get('/profiles/graph.svg', status=200)
    if not res.content_type == 'image/svg+xml' and res.json.get('status_code') == 404:
        # graphviz is probably not installed
        msg = res.json.get('message')
        assert msg == 'graph.svg is not available'
        # Force fail since graphviz is not installed on the system
        assert False
    assert res.content_type == 'image/svg+xml'
    assert res.text



def test_audit_item_schema_validation(testapp, snowball):
    testapp.patch_json(snowball['@id'] + '?validate=false', {'disallowed': 'errs'})
    res = testapp.get(snowball['@id'] + '@@index-data')
    errors = res.json['audit']
    errors_list = []
    for error_type in errors:
        errors_list.extend(errors[error_type])
    assert any(
        error['category'] == 'validation error' and error['name'] == 'audit_item_schema'
        for error in errors_list)


def test_audit_item_schema_upgrade_failure(testapp, snowball):
    testapp.patch_json(snowball['@id'] + '?validate=false', {'schema_version': '999'})
    res = testapp.get(snowball['@id'] + '@@index-data')
    errors = res.json['audit']
    errors_list = []
    for error_type in errors:
        errors_list.extend(errors[error_type])
    assert any(
        error['category'] == 'upgrade failure' and error['name'] == 'audit_item_schema'
        for error in errors_list)


def test_audit_item_schema_upgrade_ok(testapp, snowball):
    patch = {
        'schema_version': '1',
        'status': 'CURRENT',
    }
    testapp.patch_json(snowball['@id'] + '?validate=false', patch)
    res = testapp.get(snowball['@id'] + '@@index-data')
    # errors = [e for e in res.json['audit'] if e['name'] == 'audit_item_schema']
    # assert not errors
    errors = res.json['audit']
    errors_list = []
    for error_type in errors:
        errors_list.extend(errors[error_type])
    assert not any(error['name'] == 'audit_item_schema' for error in errors_list)


def test_audit_item_schema_upgrade_validation_failure(testapp, snowball):
    patch = {
        'schema_version': '1',
        'status': 'UNKNOWN',
    }
    testapp.patch_json(snowball['@id'] + '?validate=false', patch)
    res = testapp.get(snowball['@id'] + '@@index-data')
    errors = res.json['audit']
    errors_list = []
    for error_type in errors:
        errors_list.extend(errors[error_type])
    assert any(
        error['category'] == 'validation error: status' and error['name'] == 'audit_item_schema'
        for error in errors_list)


def test_audit_item_schema_permission(testapp, snowflake, embed_testapp):
    patch = {
        'type': 'wet',
        'status': 'deleted',
    }
    testapp.patch_json(snowflake['@id'], patch)
    res = embed_testapp.get(snowflake['@id'] + '/@@audit-self')
    errors_list = res.json['audit']
    assert not any(error['name'] == 'audit_item_schema' for error in errors_list)



import pytest


@pytest.fixture
def award():
    return{
        'name': 'ENCODE2',
    }


@pytest.fixture
def award_1(award):
    item = award.copy()
    item.update({
        'schema_version': '1',
        'rfa': "ENCODE2"
    })
    return item


@pytest.fixture
def award_bad(award):
    item = award.copy()
    item.update({
        'schema_version': '1',
        'rfa': "ENCODE2",
        'status': "Not a status"
    })
    return item


def test_award_upgrade(upgrader, award_1):
    value = upgrader.upgrade('award', award_1, target_version='2')
    assert value['schema_version'] == '2'
    assert value['status'] == 'disabled'


def test_award_upgrade_encode3(upgrader, award_1):
    award_1['rfa'] = 'ENCODE3'
    value = upgrader.upgrade('award', award_1, target_version='2')
    assert value['schema_version'] == '2'
    assert value['status'] == 'current'


def test_award_upgrade_url(upgrader, award_1):
    award_1['url'] = ''
    value = upgrader.upgrade('award', award_1, target_version='2')
    assert value['schema_version'] == '2'
    assert 'url' not in value


@pytest.fixture
def award_2(award_1):
    item = award_1.copy()
    item.update({
        'schema_version': '3',
        'viewing_group': 'ENCODE',
    })
    return item


def test_award_upgrade_viewing_group(upgrader, award_2):
    value = upgrader.upgrade('award', award_2, target_version='3')
    assert value['schema_version'] == '3'
    assert value['viewing_group'] == 'ENCODE'


def test_award_batch_upgrade(award_bad, testapp):
    # cannot post an invalid object
    res = testapp.post_json('/awards/', award_bad, status=422)



"""Tests the simple queue with a mocked indexer"""
import time
import uuid

from unittest import (
    TestCase,
    mock,
)

import pytest

from snovault import STORAGE
from snovault.app import main
from snovault.elasticsearch.indexer import Indexer
from snovault.elasticsearch.mpindexer import MPIndexer
from snovault.elasticsearch.interfaces import (
    APP_FACTORY,
    ELASTIC_SEARCH,
)
from snovault.elasticsearch.simple_queue import (
    SimpleUuidServer,
    SimpleUuidWorker,
)


SMALL_REQ_EMBED_WAIT = 0.01
SMALL_UUIDS_CNT = 20
SMALL_SERVE_TIMEOUT = SMALL_UUIDS_CNT * SMALL_REQ_EMBED_WAIT + 5.0
SMALL_BATCH_DIV = 3

def _get_uuids(cnt):
    """
    Return things like this 75659c39-eaea-47b7-ba26-92e9ff183e6c
    """
    uuids = set()
    while len(uuids) < cnt:
        uuids.add(str(uuid.uuid4()))
    return uuids


class MockES(object):  # pylint: disable=too-few-public-methods
    """
    Temp mock es

    There is a pytest module for this!
    https://pypi.org/project/pytest-elasticsearch/
    """
    @staticmethod
    def index(
            index=None,
            doc_type=None,
            body=None,
            id=None,
            version=None,
            version_type=None,
            request_timeout=None,
            raise_ecp=None,
        ):  # pylint: disable=unused-argument, too-many-arguments, invalid-name, redefined-builtin
        '''Fake index'''
        if raise_ecp:
            raise raise_ecp('Fake es index exception.')


class MockRegistry(dict):
    """
    Temp mock registry

    There is probably a better way to do this!
    """

    def __init__(
            self,
            batch_size,
            processes=1,
            queue_name='mockindxQ',
            queue_host=None,
            queue_port=None,
            queue_db=2,
        ):
        super().__init__()
        queue_type = 'Simple'
        self[ELASTIC_SEARCH] = MockES()
        self[STORAGE] = None
        self[APP_FACTORY] = main
        self['available_queues'] = [queue_type]
        self.settings = {
            'indexer': True,
            'snovault.elasticsearch.index': 'index-name',
            'queue_name': queue_name,
            'queue_type': queue_type,
            'queue_server': 'true',
            'queue_worker': 'true',
            'queue_worker_processes': processes,
            'queue_worker_chunk_size': 2,
            'queue_worker_batch_size': batch_size,
            'queue_worker_get_size': batch_size,
            'queue_host': queue_host,
            'queue_port': queue_port,
            'queue_db': queue_db,
        }


class MockRequest(object):  # pylint: disable=too-few-public-methods
    """
    Temp mock request

    There is probably a better way to do this!
    """
    def __init__(self, embed_wait=None):
        self.embeded_uuids = []
        self._embed_wait = embed_wait
        self._embed_errors = 0

    def set_embed_errors(self, cnt):
        '''Raise a given number of errors during embeds'''
        self._embed_errors = cnt

    def embed(self, url, as_user=None):  # pylint: disable=unused-argument
        '''Fake embed'''
        doc = {
            'item_type': 'fake item type',
        }
        if self._embed_wait:
            time.sleep(self._embed_wait)
        if self._embed_errors:
            self._embed_errors -= 1
            raise ValueError('Fake request embed exception.')
        self.embeded_uuids.append(url.split('/')[1])
        return doc


@pytest.fixture()
def small_index_objs():
    """Simple indexing objects from small vars"""
    batch_size = SMALL_UUIDS_CNT // SMALL_BATCH_DIV
    registry = MockRegistry(batch_size)
    indexer = Indexer(registry)
    invalidated = _get_uuids(SMALL_UUIDS_CNT)
    request = MockRequest(embed_wait=SMALL_REQ_EMBED_WAIT)
    return indexer, request, invalidated


# pylint: disable=redefined-outer-name
def test_smsimp_indexinit(small_index_objs):
    """Test simple indexer init with small vars"""
    indexer, _, invalidated = small_index_objs
    assert isinstance(indexer.queue_server, SimpleUuidServer)
    assert isinstance(indexer.queue_worker, SimpleUuidWorker)
    assert len(invalidated) == SMALL_UUIDS_CNT


def test_smsimp_indexserve(small_index_objs):
    """Test simple indexer serve with small vars"""
    indexer, request, invalidated = small_index_objs
    _, errors, err_msg = indexer.serve_objects(
        request,
        invalidated,
        None,  # xmin
        snapshot_id=None,
        restart=False,
        timeout=SMALL_SERVE_TIMEOUT,
    )
    assert err_msg is None
    assert isinstance(errors, list)
    assert not errors
    assert len(request.embeded_uuids) == len(invalidated)


def test_smsimp_indextimeout(small_index_objs):
    """test simple indexer serve timeout with small vars"""
    indexer, request, invalidated = small_index_objs
    _, errors, err_msg = indexer.serve_objects(
        request,
        invalidated,
        None,  # xmin
        snapshot_id=None,
        restart=False,
        timeout=0.01,
    )
    assert err_msg == 'Indexer sleep timeout'
    assert not errors


def test_smsimp_indexrun(small_index_objs):
    """test simple indexer run  with small vars"""
    indexer, request, invalidated = small_index_objs
    request.set_embed_errors(0)
    worker_runs_expected = SMALL_BATCH_DIV
    if SMALL_UUIDS_CNT % SMALL_BATCH_DIV:
        worker_runs_expected += 1
    _, errors, err_msg = indexer.serve_objects(
        request,
        invalidated,
        None,  # xmin
        snapshot_id=None,
        restart=False,
        timeout=SMALL_SERVE_TIMEOUT,
    )
    assert err_msg is None
    assert not errors
    list_invalidated = list(invalidated)
    list_invalidated.sort()
    request.embeded_uuids.sort()
    assert request.embeded_uuids == list_invalidated
    print(indexer.worker_runs)
    assert len(indexer.worker_runs) == worker_runs_expected
    uuids_ran = 0
    for worker_run in indexer.worker_runs:
        uuids_ran += worker_run['uuids']
    assert uuids_ran == SMALL_UUIDS_CNT

def _test_smsimp_indexrun_emberr(small_index_objs):
    """test simple indexer run  with small vars with embed errors"""
    indexer, request, invalidated = small_index_objs
    embed_errors = 3
    request.set_embed_errors(embed_errors)
    worker_runs_expected = SMALL_BATCH_DIV
    if SMALL_UUIDS_CNT % SMALL_BATCH_DIV:
        worker_runs_expected += 1
    _, errors, err_msg = indexer.serve_objects(
        request,
        invalidated,
        None,  # xmin
        snapshot_id=None,
        restart=False,
        timeout=SMALL_SERVE_TIMEOUT,
    )
    assert err_msg is None
    assert len(errors) == embed_errors
    list_invalidated = list(invalidated)
    list_invalidated.sort()
    request.embeded_uuids.sort()
    assert len(indexer.worker_runs) == worker_runs_expected
    uuids_ran = 0
    for worker_run in indexer.worker_runs:
        uuids_ran += worker_run['uuids']
    assert uuids_ran == SMALL_UUIDS_CNT


def test_simple_mpindexinit():
    """test simple mpindexer"""
    batch_size = SMALL_UUIDS_CNT // SMALL_BATCH_DIV
    registry = MockRegistry(batch_size)
    processes = registry.settings['queue_worker_processes']
    # pylint: disable=unused-variable
    mpindexer = MPIndexer(registry, processes=processes)
    invalidated = _get_uuids(SMALL_UUIDS_CNT)
    request = MockRequest(embed_wait=SMALL_REQ_EMBED_WAIT)
    assert True


class TestIndexer(TestCase):
    """Test Indexer in indexer.py"""
    @classmethod
    def setUpClass(cls):
        # Request
        cls.embed_wait = 0.01
        # Server Objedcts
        cls.invalidated = _get_uuids(2500)
        cls.xmin = None
        cls.snapshot_id = 'Some-Snapshot-Id'
        cls.restart = False
        # Indexer
        batch_size = 100
        processes = 1
        registry = MockRegistry(batch_size, processes=processes)
        cls.indexer = Indexer(registry)

    def test_init(self):
        """Test Indexer Init"""
        self.assertIsInstance(self.indexer.queue_server, SimpleUuidServer)
        self.assertIsInstance(self.indexer.queue_worker, SimpleUuidWorker)

    def test_serve_objects_no_uuids(self):
        """Test serve objects with no_uuids"""
        expected_err_msg = (
            'Cannot initialize indexing process: '
            'No uuids given to Indexer.serve_objects'
        )
        request = MockRequest(embed_wait=self.embed_wait)
        uuids = []
        _, errors, err_msg = self.indexer.serve_objects(
            request,
            uuids,
            self.xmin,
            snapshot_id=self.snapshot_id,
            restart=self.restart,
            timeout=100,
        )
        self.assertEqual(err_msg, expected_err_msg)
        self.assertListEqual(errors, [])

    def test_serve_objects_noload(self):
        """Test serve objects when fails to load any uuids"""
        uuids = self.invalidated
        org_load_uuids = self.indexer.queue_server.load_uuids
        self.indexer.queue_server.load_uuids = mock.MagicMock(return_value=None)
        expected_err_msg = 'Uuids given to Indexer.serve_objects failed to load'
        request = MockRequest(embed_wait=self.embed_wait)
        _, errors, err_msg = self.indexer.serve_objects(
            request,
            uuids,
            self.xmin,
            snapshot_id=self.snapshot_id,
            restart=self.restart,
            timeout=100,
        )
        self.assertEqual(err_msg, expected_err_msg)
        self.assertListEqual(errors, [])
        self.indexer.queue_server.load_uuids = org_load_uuids

    def test_serve_objects_partload(self):
        """Test serve objects when fails to load all uuids"""
        uuids_loaded_cnt = 1
        uuids = self.invalidated
        org_load_uuids = self.indexer.queue_server.load_uuids
        self.indexer.queue_server.load_uuids = mock.MagicMock(return_value=uuids_loaded_cnt)
        expected_err_msg = (
            'Uuids given to Indexer.serve_objects '
            'failed to all load. {} of {} only'.format(
                uuids_loaded_cnt,
                len(uuids),
            )
        )
        request = MockRequest(embed_wait=self.embed_wait)
        _, errors, err_msg = self.indexer.serve_objects(
            request,
            uuids,
            self.xmin,
            snapshot_id=self.snapshot_id,
            restart=self.restart,
            timeout=100,
        )
        self.assertEqual(err_msg, expected_err_msg)
        self.assertListEqual(errors, [])
        self.indexer.queue_server.load_uuids = org_load_uuids

    def test_serve_objects_isindexing(self):
        """Test serve objects when already indexing"""
        uuids = self.invalidated
        org_is_indexing = self.indexer.queue_server.is_indexing
        self.indexer.queue_server.is_indexing = mock.MagicMock(return_value=True)
        expected_err_msg = 'Cannot initialize indexing process: Already Indexing'
        request = MockRequest(embed_wait=self.embed_wait)
        _, errors, err_msg = self.indexer.serve_objects(
            request,
            uuids,
            self.xmin,
            snapshot_id=self.snapshot_id,
            restart=self.restart,
            timeout=100,
        )
        self.assertEqual(err_msg, expected_err_msg)
        self.assertListEqual(errors, [])
        self.indexer.queue_server.is_indexing = org_is_indexing

    def test_serve_objects_noworker(self):
        """
        Test serve objects when no work.  Should timeout for simple queue

        Since worker cannot finish we manually reset
            - self.indexer.queue_server._uuids = []
        """
        uuids = self.invalidated
        org_queue_worker = self.indexer.queue_worker
        self.indexer.queue_worker = None
        expected_err_msg = 'Indexer sleep timeout'
        request = MockRequest(embed_wait=self.embed_wait)
        _, errors, err_msg = self.indexer.serve_objects(
            request,
            uuids,
            self.xmin,
            snapshot_id=self.snapshot_id,
            restart=self.restart,
            timeout=1,
        )
        # pylint: disable=protected-access
        self.indexer.queue_server._uuids = []
        self.assertEqual(err_msg, expected_err_msg)
        self.assertListEqual(errors, [])
        self.indexer.queue_worker = org_queue_worker

    def test_serve_objects_timeout(self):
        """
        Test serve objects timeout

        Since worker cannot finish we manually reset
            - self.indexer.queue_server._uuids = []
            - serve objects can only timeout after worker finishes
        """
        uuids = self.invalidated
        embed_wait = 0.01
        expected_err_msg = 'Indexer sleep timeout'
        request = MockRequest(embed_wait=embed_wait)
        _, errors, err_msg = self.indexer.serve_objects(
            request,
            uuids,
            self.xmin,
            snapshot_id=self.snapshot_id,
            restart=self.restart,
            timeout=0.1,
        )
        # pylint: disable=protected-access
        self.indexer.queue_server._uuids = []
        self.assertEqual(err_msg, expected_err_msg)
        self.assertListEqual(errors, [])

    def test_serve_objects_batcherrs(self):
        """
        Test serve objects returns batch errors
        """
        uuids = list(self.invalidated)[0:100]
        embed_wait = 0.01
        return_errors = ['e1', 'e2', 'e3']
        org_pop_errors = self.indexer.queue_server.pop_errors
        self.indexer.queue_server.pop_errors = mock.MagicMock(return_value=return_errors)
        request = MockRequest(embed_wait=embed_wait)
        _, errors, err_msg = self.indexer.serve_objects(
            request,
            uuids,
            self.xmin,
            snapshot_id=self.snapshot_id,
            restart=self.restart,
            timeout=10,
        )
        self.indexer.queue_server.pop_errors = org_pop_errors
        # pylint: disable=protected-access
        self.indexer.queue_server._uuids = []
        self.assertIsNone(err_msg)
        errors.sort()
        return_errors.sort()
        self.assertListEqual(errors, return_errors)

    def test_serve_objects_run(self):
        """
        Test serve objects run
        """
        uuids = self.invalidated
        embed_wait = 0.001
        batch_size = 1024
        self.indexer.queue_worker.queue_options['batch_size'] = batch_size
        request = MockRequest(embed_wait=embed_wait)
        _, errors, err_msg = self.indexer.serve_objects(
            request,
            uuids,
            self.xmin,
            snapshot_id=self.snapshot_id,
            restart=self.restart,
            timeout=100,
        )
        self.assertListEqual(errors, [])
        self.assertIsNone(err_msg)

    def test_serve_objects_runerrs(self):
        """
        Test serve objects run with errors
        """
        uuids = self.invalidated
        embed_wait = 0.001
        batch_size = 1024
        self.indexer.queue_worker.queue_options['batch_size'] = batch_size
        request = MockRequest(embed_wait=embed_wait)
        request.set_embed_errors(10)
        _, errors, err_msg = self.indexer.serve_objects(
            request,
            uuids,
            self.xmin,
            snapshot_id=self.snapshot_id,
            restart=self.restart,
            timeout=100,
        )
        self.assertEqual(len(errors), 10)
        self.assertIsNone(err_msg)

    def test_serve_objects_twiceerrs(self):
        """
        Test serve objects running twice with first run having errors
        """
        uuids = self.invalidated
        embed_wait = 0.001
        batch_size = 1024
        self.indexer.queue_worker.queue_options['batch_size'] = batch_size
        # Error Run
        request = MockRequest(embed_wait=embed_wait)
        request.set_embed_errors(10)
        _, errors, err_msg = self.indexer.serve_objects(
            request,
            uuids,
            self.xmin,
            snapshot_id=self.snapshot_id,
            restart=self.restart,
            timeout=100,
        )
        self.assertEqual(len(errors), 10)
        self.assertIsNone(err_msg)
        # Clean Run
        request = MockRequest(embed_wait=embed_wait)
        _, errors, err_msg = self.indexer.serve_objects(
            request,
            uuids,
            self.xmin,
            snapshot_id=self.snapshot_id,
            restart=self.restart,
            timeout=100,
        )
        self.assertListEqual(errors, [])
        self.assertIsNone(err_msg)



"""csv compatible interface for xlrd sheets
"""

import csv
import datetime
import os.path
import xlrd
import zipfile


def cell_value(cell, datemode):
    ctype = cell.ctype
    value = cell.value

    if ctype == xlrd.XL_CELL_ERROR:
        raise ValueError(repr(cell), 'cell error')

    elif ctype == xlrd.XL_CELL_BOOLEAN:
        return str(value).upper()

    elif ctype == xlrd.XL_CELL_NUMBER:
        if value.is_integer():
            value = int(value)
        return str(value)

    elif ctype == xlrd.XL_CELL_DATE:
        value = xlrd.xldate_as_tuple(value, datemode)
        if value[3:] == (0, 0, 0):
            return datetime.date(*value[:3]).isoformat()
        else:
            return datetime.datetime(*value).isoformat()

    elif ctype in (xlrd.XL_CELL_TEXT, xlrd.XL_CELL_EMPTY, xlrd.XL_CELL_BLANK):
        return value

    raise ValueError(repr(cell), 'unknown cell type')



def reader(stream, sheetname=None):
    """ Read named sheet or first and only sheet from xlsx file
    """
    book = xlrd.open_workbook(file_contents=stream.read())
    if sheetname is None:
        sheet, = book.sheets()        
    else:
        try:
            sheet = book.sheet_by_name(sheetname)
        except xlrd.XLRDError:
            return

    datemode = sheet.book.datemode
    for index in range(sheet.nrows):
        yield [cell_value(cell, datemode) for cell in sheet.row(index)]


class DictReader:
    # Adapted from http://hg.python.org/cpython/file/2.7/Lib/csv.py
    def __init__(self, stream, fieldnames=None, restkey=None, restval=None,
                 *args, **kwds):
        self._fieldnames = fieldnames   # list of keys for the dict
        self.restkey = restkey          # key to catch long rows
        self.restval = restval          # default value for short rows
        self.reader = reader(stream, *args, **kwds)
        self.line_num = 0

    def __iter__(self):
        return self

    @property
    def fieldnames(self):
        if self._fieldnames is None:
            try:
                self._fieldnames = next(self.reader)
            except StopIteration:
                pass
            else:
                self.line_num += 1
        #self.line_num = self.reader.line_num
        return self._fieldnames

    @fieldnames.setter
    def fieldnames(self, value):
        self._fieldnames = value

    def __next__(self):
        if self._fieldnames is None:
            # Used only for its side effect.
            self.fieldnames
        row = next(self.reader)
        self.line_num += 1

        # unlike the basic reader, we prefer not to return blanks,
        # because we will typically wind up with a dict full of None
        # values
        while row == []:
            row = next(self.reader)
        d = dict(zip(self.fieldnames, row))
        lf = len(self.fieldnames)
        lr = len(row)
        if lf < lr:
            d[self.restkey] = row[lf:]
        elif lf > lr:
            for key in self.fieldnames[lr:]:
                d[key] = self.restval
        return d


def zipfile_to_csv(zipfilename, outpath, ext='.csv', dialect='excel', **fmtparams):
    """ For Google Drive download zips
    """
    zf = ZipFile(zipfilename)
    for name in zf.namelist():
        subpath, entry_ext = os.path.splitext(name)
        if entry_ext.lower() != '.xlsx':
            continue
        f = zf.open(name)
        book = xlrd.open_workbook(file_contents=f.read())
        sheet, = book.sheets()  # Only single worksheet books are handled
        csvfile = open(os.path.join(outpath, subpath + ext), 'w')
        wr = csv.writer(csvfile, dialect=dialect, **fmtparams)
        wr.writerows(list(reader(sheet)))




def test_batch_upgrade(testapp, invalid_snowball):

    assert invalid_snowball['status'] == 'FLYBARGS'
    upgrade = testapp.post_json(
        '/batch_upgrade', {'batch': [invalid_snowball['uuid']]}, status=200)

    res = testapp.get(invalid_snowball['@id'])
    assert res.json['schema_version'] == '2'
    err_msg = (
        "Validation failure: /snowball/{uuid}\n"
        "status: 'FLYBARGS' is not one of "
        "['proposed', 'started', 'submitted', 'ready for review', "
        "'deleted', 'released', 'revoked', 'archived', 'replaced']"
        "".format(
            uuid=invalid_snowball['uuid'],
        )
    )
    assert upgrade.json['results'][0] == [
        'snowball',
        invalid_snowball['uuid'],
        True,  # updated
        True,   # errors
        err_msg,
    ]


def test_batch_upgrade_multi(testapp, invalid_snowball, invalid_snowball2):

    assert invalid_snowball2['alternate_accessions'] == ['THE THING THAT SHOULD NOT BE']
    upgrade = testapp.post_json(
        '/batch_upgrade', {'batch': [invalid_snowball['uuid'], invalid_snowball2['uuid']]}, status=200)

    assert len(upgrade.json['results']) == 2
    err_msg = (
        "Validation failure: /snowball/{uuid}\n"
        "status: 'FLYBARGS' is not one of "
        "['proposed', 'started', 'submitted', 'ready for review', "
        "'deleted', 'released', 'revoked', 'archived', 'replaced']"
        "".format(
            uuid=invalid_snowball['uuid'],
        )
    )
    assert upgrade.json['results'][0] == [
        'snowball',
        invalid_snowball['uuid'],
        True,  # updated
        True,   # errors
        err_msg,
    ]


def test_batch_upgrade_regex(testapp, invalid_award):
    assert invalid_award['name'] == 'award names cannot have spaces'
    upgrade = testapp.post_json('/batch_upgrade', {'batch': [invalid_award['uuid']]}, status=200)
    err_msg = (
        "Validation failure: /award/{uuid}\n"
        "name: 'award names cannot have spaces' "
        "does not match '^[A-Za-z0-9\\\\-]+$'"
        "".format(
            uuid=invalid_award['uuid'],
        )
    )
    assert upgrade.json['results'][0] == [
        'award',
        invalid_award['uuid'],
        True,  # updated
        True,   # errors
        err_msg,
    ]



"""\
Also run this when the links or keys are changed in the schema.

Example:

Demo
    batchupgrade development.ini --app-name app
Production or Release Candidate
    batchupgrade production.ini --app-name app --processes 16 --batchsize 1000

"""
import logging
import time
import transaction

from copy import deepcopy
from itertools import groupby

from pyramid import paster
from pyramid.traversal import find_resource
from pyramid.view import view_config

from snovault import (
    CONNECTION,
    STORAGE,
    UPGRADER,
)
from snovault.schema_utils import validate


BATCH_UPGRADE_LOG = logging.getLogger('snovault.batchupgrade')
EPILOG = __doc__
testapp = None


def includeme(config):
    config.add_route('batch_upgrade', '/batch_upgrade')
    config.scan(__name__)


def update_item(storage, context):
    target_version = context.type_info.schema_version
    current_version = context.properties.get('schema_version', '')
    update = False
    errors = []
    properties = context.properties
    if target_version is None or current_version == target_version:
        unique_keys = context.unique_keys(properties)
        links = context.links(properties)
        keys_add, keys_remove = storage._update_keys(context.model, unique_keys)
        if keys_add or keys_remove:
            update = True
        rels_add, rels_remove = storage._update_rels(context.model, links)
        if rels_add or rels_remove:
            update = True
    else:
        properties = deepcopy(properties)
        upgrader = context.registry[UPGRADER]
        properties = upgrader.upgrade(
            context.type_info.name, properties, current_version, target_version,
            context=context, registry=context.registry)
        if 'schema_version' in properties:
            del properties['schema_version']
        schema = context.type_info.schema
        properties['uuid'] = str(context.uuid)
        validated, errors = validate(schema, properties, properties)
        # Do not send modification events to skip indexing
        context.update(validated)
        update = True
    return update, errors


@view_config(route_name='batch_upgrade', request_method='POST', permission='import_items')
def batch_upgrade(request):
    request.datastore = 'database'
    transaction.get().setExtendedInfo('upgrade', True)
    batch = request.json['batch']
    root = request.root
    storage = request.registry[STORAGE].write
    session = storage.DBSession()
    results = []
    for uuid in batch:
        item_type = None
        update = False
        error = False
        error_msg = ''
        sp = session.begin_nested()
        try:
            item = find_resource(root, uuid)
            item_type = item.type_info.item_type
            update, errors = update_item(storage, item)
        except Exception as ecp:
            error_msg = "Exception: {} updating: /{}/{}".format(
                ecp,
                item_type,
                uuid,
            )
            sp.rollback()
            error = True
        else:
            if errors:
                # redmine 5161 sometimes error.path has an int
                errortext = [
                    '%s: %s' % ('/'.join([str(x) or '<root>' for x in error.path]), error.message)
                    for error in errors
                ]
                error_msg = "Validation failure: /{}/{}\n{}".format(
                    item_type,
                    uuid,
                    '\n'.join(errortext),
                )
                sp.rollback()
                error = True
            else:
                sp.commit()
        results.append((item_type, uuid, update, error, error_msg))
    return {'results': results}


def _pool_batch_results(iterable, n=1):
    l = len(iterable)
    for ndx in range(0, l, n):
        yield iterable[ndx:min(ndx+n, l)]


def _pool_initializer(*args, **kw):
    global testapp
    testapp = _internal_app(*args, **kw)


def _pool_worker(batch):
    res = testapp.post_json('/batch_upgrade', {'batch': batch})
    return res.json


def _pool_chunksize(uuid_count, chunksize, processes):
    chunkiness = int((uuid_count - 1) / processes) + 1
    if chunkiness > chunksize:
        return chunksize
    return chunkiness


def _run_pool(uuids, args):
    from multiprocessing import get_context
    from multiprocessing.pool import Pool
    transaction.abort()
    pool = Pool(
        processes=args.processes,
        initializer=_pool_initializer,
        initargs=(args.config_uri, args.app_name, args.username),
        context=get_context('forkserver'),
        maxtasksperchild=args.maxtasksperchild,
    )
    est_loops = int(len(uuids) / args.batchsize)
    all_results = []
    try:
        pool_gen = pool.imap_unordered(
            _pool_worker,
            _pool_batch_results(uuids, args.batchsize),
            chunksize=args.chunksize,
        )
        for loop, result in enumerate(pool_gen, 1):
            results = result['results']
            error_msgs = [
                error_msg
                for _, _, _, _, error_msg in results
                if error_msg
            ]
            updated_cnt = sum(update for _, _, update, _, _ in results)
            log_msg = "{} of ~{} Batch: Updated {} of {} (errors {})".format(
                loop,
                est_loops,
                updated_cnt,
                len(results),
                len(error_msgs),
            )
            BATCH_UPGRADE_LOG.info(log_msg)
            for error_msg in error_msgs:
                BATCH_UPGRADE_LOG.error("\t%s", error_msg)
            all_results.extend(results)
    finally:
        pool.terminate()
        pool.join()
    return all_results


def _summarize_results(all_results, runtime_str=None, verbose=False):

    def result_item_type(result):
        return result[0] or ''

    error_logs = []
    updated_logs = []
    all_logs = []
    for item_type, results in groupby(
            sorted(all_results, key=result_item_type), key=result_item_type):
        results = list(results)
        errors_cnt = sum(error for _, _, _, error, _ in results)
        updated_cnt = sum(update for _, _, update, _, _ in results)
        log_message = "Collection {}: Updated {} of {} (errors {})".format(
            item_type,
            updated_cnt,
            len(results),
            errors_cnt,
        )
        if errors_cnt:
            error_logs.append(log_message)
        if updated_cnt:
            updated_logs.append(log_message)
        all_logs.append(log_message)
    BATCH_UPGRADE_LOG.info('Upgrade Summary')
    if verbose:
        for log_msg in all_logs:
            BATCH_UPGRADE_LOG.info(log_msg)
    BATCH_UPGRADE_LOG.info('Sum updated: %d' % len(updated_logs))
    if updated_logs:
        for log_msg in updated_logs:
            BATCH_UPGRADE_LOG.info(log_msg)
    BATCH_UPGRADE_LOG.info('Sum errors: %d' % len(error_logs))
    if error_logs:
        for log_msg in error_logs:
            BATCH_UPGRADE_LOG.error(log_msg)
    if runtime_str:
        BATCH_UPGRADE_LOG.info("Run Time: %s" % runtime_str)


def _internal_app(configfile, app_name=None, username=None):
    from webtest import TestApp
    from pyramid import paster
    app = paster.get_app(configfile, app_name)
    if not username:
        username = 'UPGRADE'
    environ = {
        'HTTP_ACCEPT': 'application/json',
        'REMOTE_USER': username,
    }
    return TestApp(app, environ)


def main():
    args = _parse_args()
    # Setup Logger
    paster.setup_logging(args.config_uri)
    logging.getLogger('snovault').setLevel(logging.INFO)
    # Get Uuids
    testapp = _internal_app(args.config_uri, app_name=args.app_name, username=args.username)
    connection = testapp.app.registry[CONNECTION]
    uuids = [str(uuid) for uuid in connection.__iter__(*args.item_types)]
    if uuids:
        log_msg = "Start Upgrade with {} items: {}, {}, {}, {}".format(
            len(uuids),
            args.batchsize,
            args.chunksize,
            args.processes,
            args.maxtasksperchild,
        )
        BATCH_UPGRADE_LOG.info(log_msg)
        pool_start = time.time()
        all_results = _run_pool(uuids, args)
        runtime_mins_str = "{:0.2f} minutes".format(
            (time.time() - pool_start) / 60
        )
        BATCH_UPGRADE_LOG.info('End Upgrade')
        _summarize_results(all_results, runtime_str=runtime_mins_str, verbose=args.verbose)
    else:
        BATCH_UPGRADE_LOG.warning('No uuids to upgrade.')


def _parse_args():
    import argparse
    parser = argparse.ArgumentParser(
        description="Batch upgrade content items.", epilog=EPILOG,
        formatter_class=argparse.RawDescriptionHelpFormatter,
    )
    parser.add_argument('config_uri', help="path to configfile")
    parser.add_argument('--app-name', help="Pyramid app name in configfile")
    parser.add_argument('--batchsize', type=int, default=50)
    parser.add_argument('--chunksize', type=int, default=1)
    parser.add_argument('--item-types', action='append', default=[])
    parser.add_argument('--maxtasksperchild', type=int, default=1)
    parser.add_argument('--processes', type=int, default=2)
    parser.add_argument('--username')
    parser.add_argument('-v', '--verbose', action='store_true')
    args = parser.parse_args()
    return args


if __name__ == '__main__':
    main()



from snovault import upgrade_step

''' Note these are not relevant to anything beyond testing upgrader '''


@upgrade_step('award', '', '2')
def award_0_2(value, system):
    # Sample upgrades with tests

    rfa_mapping = ['ENCODE2', 'ENCODE2-Mouse']
    if value['rfa'] in rfa_mapping:
        value['status'] = 'disabled'
    else:
        value['status'] = 'current'

    if 'url' in value:
        if value['url'] == '':
            del value['url']


@upgrade_step('award', '2', '3')
def award_2_3(value, system):

    if value['viewing_group'] == 'ENCODE3':
        value['viewing_group'] = 'ENCODE'



import sys
from future.utils import (
    raise_with_traceback,
    itervalues,
)
from itertools import islice
from past.builtins import basestring
from pyramid.exceptions import PredicateMismatch
from pyramid.httpexceptions import HTTPNotFound
from pyramid.settings import asbool
from pyramid.view import (
    render_view_to_response,
    view_config,
)
from urllib.parse import urlencode
from snosearch.fields import AllResponseField
from snosearch.fields import CollectionSearchWithFacetsResponseField
from snosearch.fields import ColumnsResponseField
from snosearch.fields import CollectionClearFiltersResponseField
from snosearch.fields import FiltersResponseField
from snosearch.fields import IDResponseField
from snosearch.fields import NotificationResponseField
from snosearch.parsers import ParamsParser
from snosearch.parsers import QueryString
from snosearch.responses import FieldedResponse
from .calculated import calculate_properties
from .calculated import calculate_select_properties
from .calculated import calculate_filtered_properties
from .calculated import _should_render_property
from .etag import etag_tid
from .interfaces import CONNECTION
from .elasticsearch.interfaces import ELASTIC_SEARCH
from .resources import (
    AbstractCollection,
    Item,
    Root,
)
from .util import expand_path


def includeme(config):
    config.scan(__name__)


def remove_item_keys(item, request):
    """Remove keys specified in query from item"""
    removed_fields = request.params.getall('remove')
    for field in removed_fields:
        if field in item:
            del item[field]
    return item


@view_config(context=AbstractCollection, permission='list', request_method='GET',
             name='listing_db')
def collection_view_listing_db(context, request):
    result = {}

    frame = request.params.get('frame', 'columns')

    limit = request.params.get('limit', 25)
    if limit in ('', 'all'):
        limit = None
    if limit is not None:
        try:
            limit = int(limit)
        except ValueError:
            limit = 25

    items = (
        item for item in itervalues(context)
        if request.has_permission('view', item)
    )

    if limit is not None:
        items = islice(items, limit)

    result['@graph'] = [
        remove_item_keys(request.embed(request.resource_path(item, '@@' + frame)), request)
        for item in items
    ]

    if limit is not None and len(result['@graph']) == limit:
        params = [(k, v) for k, v in request.params.items() if k != 'limit']
        params.append(('limit', 'all'))
        result['all'] = '%s?%s' % (request.resource_path(context), urlencode(params))

    return result


@view_config(context=AbstractCollection, permission='list', request_method='GET', name='listing')
def collection_view_listing_es(context, request):
    if not hasattr(request, 'datastore') or request.datastore != ELASTIC_SEARCH:
        return collection_view_listing_db(context, request)
    fr = FieldedResponse(
        _meta={
            'params_parser': ParamsParser(request)
        },
        response_fields=[
            IDResponseField(),
            CollectionSearchWithFacetsResponseField(),
            AllResponseField(),
            NotificationResponseField(),
            FiltersResponseField(),
            CollectionClearFiltersResponseField(),
            ColumnsResponseField()
        ]
    )
    return fr.render()


@view_config(context=Root, request_method='GET', name='page')
def home(context, request):
    properties = request.embed(request.resource_path(context), '@@object')
    calculated = calculate_properties(context, request, properties, category='page')
    properties.update(calculated)
    return properties


@view_config(context=Root, request_method='GET', name='object')
@view_config(context=AbstractCollection, permission='list', request_method='GET', name='object')
def collection_view_object(context, request):
    properties = context.__json__(request)
    calculated = calculate_properties(context, request, properties)
    properties.update(calculated)
    return properties


@view_config(context=AbstractCollection, permission='list', request_method='GET', name='page')
def collection_list(context, request):
    path = request.resource_path(context)
    properties = request.embed(path, '@@object')
    calculated = calculate_properties(context, request, properties, category='page')
    properties.update(calculated)

    if request.query_string:
        properties['@id'] += '?' + request.query_string

    result = request.embed(path, '@@listing?' + request.query_string, as_user=True)
    result.update(properties)
    return result


@view_config(context=Root, request_method='GET')
@view_config(context=AbstractCollection, permission='list', request_method='GET')
@view_config(context=Item, permission='view', request_method='GET')
def item_view(context, request):
    frame = request.params.get('frame', 'page')
    if getattr(request, '__parent__', None) is None:
        # We need the response headers from non subrequests
        try:
            response = render_view_to_response(context, request, name=frame)
        except PredicateMismatch:
            # Avoid this view emitting PredicateMismatch
            exc_class, exc, tb = sys.exc_info()
            exc.__class__ = HTTPNotFound
            raise_with_traceback(exc, tb)
        else:
            if response is None:
                raise HTTPNotFound('?frame=' + frame)
            return response
    path = request.resource_path(context, '@@' + frame)
    if request.query_string:
        path += '?' + request.query_string
    return request.embed(path, as_user=True)


def item_links(context, request):
    # This works from the schema rather than the links table
    # so that upgrade on GET can work.
    properties = context.__json__(request)
    for path in context.type_info.schema_links:
        uuid_to_path(request, properties, path)
    return properties


def uuid_to_path(request, obj, path):
    if isinstance(path, basestring):
        path = path.split('.')
    if not path:
        return
    name = path[0]
    remaining = path[1:]
    value = obj.get(name, None)
    if value is None:
        return
    if remaining:
        if isinstance(value, list):
            for v in value:
                uuid_to_path(request, v, remaining)
        else:
            uuid_to_path(request, value, remaining)
        return
    conn = request.registry[CONNECTION]
    if isinstance(value, list):
        obj[name] = [
            request.resource_path(conn[v])
            for v in value
        ]
    else:
        obj[name] = request.resource_path(conn[value])


@view_config(context=Item, permission='view', request_method='GET',
             name='object')
def item_view_object(context, request):
    """ Render json structure

    1. Fetch stored properties, possibly upgrading.
    2. Link canonicalization (overwriting uuids.)
    3. Calculated properties (including reverse links.)
    """
    properties = item_links(context, request)
    if not asbool(request.params.get('skip_calculated')):
        calculated = calculate_properties(context, request, properties)
        properties.update(calculated)
    return properties


@view_config(
    context=Item,
    permission='view',
    request_method='GET',
    name='object_with_select_calculated_properties'
)
def item_view_object_with_select_calculated_properties(context, request):
    properties = item_links(context, request)
    qs = QueryString(request)
    select_properties = qs.param_values_to_list(
        params=qs.get_field_filters()
    )
    calculated = calculate_select_properties(
        context,
        request,
        ns=properties,
        select_properties=select_properties
        )
    properties.update(calculated)
    return properties


@view_config(
    context=Item,
    permission='view',
    request_method='GET',
    name='filtered_object'
)
def item_view_filtered_object(context, request):
    properties = item_links(context, request)
    qs = QueryString(request)
    include = qs.param_values_to_list(
        params=qs.get_key_filters(
            key='include'
        )
    ) or None
    exclude = qs.param_values_to_list(
        params=qs.get_key_filters(
            key='exclude'
        )
    ) or None
    calculated = calculate_filtered_properties(
        context,
        request,
        ns=properties,
        include=include,
        exclude=exclude,
    )
    properties.update(calculated)
    return {
        k: v
        for k, v in properties.items()
        if _should_render_property(include, exclude, k)
    }


@view_config(context=Item, permission='view', request_method='GET',
             name='embedded')
def item_view_embedded(context, request):
    item_path = request.resource_path(context)
    properties = request.embed(item_path, '@@object')
    for path in context.embedded:
        expand_path(request, properties, path)
    for path in context.embedded_with_frame:
        path.expand(request, properties)
    return properties


@view_config(context=Item, permission='view', request_method='GET',
             name='page')
def item_view_page(context, request):
    item_path = request.resource_path(context)
    properties = request.embed(item_path, '@@embedded')
    calculated = calculate_properties(context, request, properties, category='page')
    properties.update(calculated)
    return properties


@view_config(context=Item, permission='expand', request_method='GET',
             name='expand')
def item_view_expand(context, request):
    path = request.resource_path(context)
    properties = request.embed(path, '@@object')
    for path in request.params.getall('expand'):
        expand_path(request, properties, path)
    return properties


def expand_column(request, obj, subset, path):
    if isinstance(path, basestring):
        path = path.split('.')
    if not path:
        return
    name = path[0]
    remaining = path[1:]
    value = obj.get(name, None)
    if value is None:
        return
    if not remaining:
        subset[name] = value
        return
    if isinstance(value, list):
        if name not in subset:
            subset[name] = [{} for i in range(len(value))]
        for index, member in enumerate(value):
            if not isinstance(member, dict):
                member = request.embed(member, '@@object')
            expand_column(request, member, subset[name][index], remaining)
    else:
        if name not in subset:
            subset[name] = {}
        if not isinstance(value, dict):
            value = request.embed(value, '@@object')
        expand_column(request, value, subset[name], remaining)


@view_config(context=Item, permission='view', request_method='GET',
             name='columns')
def item_view_columns(context, request):
    path = request.resource_path(context)
    properties = request.embed(path, '@@object')
    if context.schema is None or 'columns' not in context.schema:
        return properties

    subset = {
        '@id': properties['@id'],
        '@type': properties['@type'],
    }

    for column in context.schema['columns']:
        path = column.split('.')
        if path[-1] == 'length':
            path.pop()
        if path:
            expand_column(request, properties, subset, path)

    return subset


@view_config(context=Item, permission='view_raw', request_method='GET',
             name='raw')
def item_view_raw(context, request):
    if asbool(request.params.get('upgrade', True)):
        return context.upgrade_properties()
    return context.properties


@view_config(context=Item, permission='edit', request_method='GET',
             name='edit', decorator=etag_tid)
def item_view_edit(context, request):
    conn = request.registry[CONNECTION]
    properties = item_links(context, request)
    schema_rev_links = context.type_info.schema_rev_links

    for propname in schema_rev_links:
        properties[propname] = sorted(
            request.resource_path(child)
            for child in (
                conn.get_by_uuid(uuid) for uuid in context.get_rev_links(propname)
            ) if request.has_permission('visible_for_edit', child)
        )

    return properties


@view_config(context=Item, permission='edit', request_method='GET',
             name='form', decorator=etag_tid)
def item_view_form(context, request):
    properties = item_view_edit(context, request)
    properties['@type'] = context.jsonld_type()
    return properties



from past.builtins import basestring
from pyramid.threadlocal import manager as threadlocal_manager
from snovault.interfaces import ROOT
from urllib.parse import urlencode


def includeme(config):
    config.add_request_method(select_distinct_values)


def get_root_request():
    if threadlocal_manager.stack:
        return threadlocal_manager.stack[0]['request']


def ensurelist(value):
    if isinstance(value, basestring):
        return [value]
    return value


def simple_path_ids(obj, path):
    if isinstance(path, basestring):
        path = path.split('.')
    if not path:
        yield obj
        return
    name = path[0]
    remaining = path[1:]
    value = obj.get(name, None)
    if value is None:
        return
    if not isinstance(value, list):
        value = [value]
    for member in value:
        for result in simple_path_ids(member, remaining):
            yield result


def expand_path(request, obj, path):
    if isinstance(path, basestring):
        path = path.split('.')
    if not path:
        return
    name = path[0]
    remaining = path[1:]
    value = obj.get(name, None)
    if value is None:
        return
    if isinstance(value, list):
        for index, member in enumerate(value):
            if not isinstance(member, dict):
                member = value[index] = request.embed(member, '@@object')
            expand_path(request, member, remaining)
    else:
        if not isinstance(value, dict):
            value = obj[name] = request.embed(value, '@@object')
        expand_path(request, value, remaining)


def _get_calculated_properties_from_paths(request, paths):
    root = request.registry[ROOT]
    calculated_fields = set()
    # Expect paths in form ['/item1/identifier/', '/item2/identifier/', ...].
    item_types = {
        root.collections.get(p.split('/')[1]).type_info.item_type
        for p in paths
        if len(p.split('/')) == 4 and p.split('/')[1] in root.collections
    }
    for item_type in item_types:
        item_cls = request.registry['types'].by_item_type.get(item_type).factory
        calculated_fields.update(
            root.collections.get(item_type).type_info.calculated_properties.props_for(item_cls).keys()
        )
    return list(calculated_fields)


def select_distinct_values(request, value_path, *from_paths):
    if isinstance(value_path, basestring):
        value_path = value_path.split('.')
    values = from_paths
    for name in value_path:
        calculated_properties = _get_calculated_properties_from_paths(request, values)
        # Don't waste time calculating properties if the field isn't calculated.
        frame = '@@object' if name in calculated_properties else '@@object?skip_calculated=true'
        objs = (request.embed(member, frame) for member in values)
        value_lists = (ensurelist(obj.get(name, [])) for obj in objs)
        values = {value for value_list in value_lists for value in value_list}
    return list(values)


def quick_deepcopy(obj):
    """Deep copy an object consisting of dicts, lists, and primitives.

    This is faster than Python's `copy.deepcopy` because it doesn't
    do bookkeeping to avoid duplicating objects in a cyclic graph.

    This is intended to work fine for data deserialized from JSON,
    but won't work for everything.
    """
    if isinstance(obj, dict):
        obj = {k: quick_deepcopy(v) for k, v in obj.items()}
    elif isinstance(obj, list):
        obj = [quick_deepcopy(v) for v in obj]
    return obj


def mutated_schema(schema, mutator):
    """Apply a change to all levels of a schema.

    Returns a new schema rather than modifying the original.
    """
    schema = mutator(schema.copy())
    if 'items' in schema:
        schema['items'] = mutated_schema(schema['items'], mutator)
    if 'properties' in schema:
        schema['properties'] = schema['properties'].copy()
        for k, v in schema['properties'].items():
            schema['properties'][k] = mutated_schema(v, mutator)
    return schema


def try_to_get_field_from_item_with_skip_calculated_first(request, field, at_id_of_item):
    item = request.embed(at_id_of_item, '@@object?skip_calculated=true')
    if field not in item:
        item = request.embed(
            at_id_of_item,
            '@@object_with_select_calculated_properties?field={}'.format(field)
        )
    return item.get(field)


def ensure_list_and_filter_none(values):
    if not isinstance(values, list):
        values = [values]
    return [
        v
        for v in values
        if v is not None
    ]


def take_one_or_return_none(values):
    if isinstance(values, list) and len(values) == 1:
        return values[0]


class Path:

    FILTER_FRAME = '@@filtered_object'

    def __init__(self, path, frame='@@object', include=None, exclude=None):
        self.path = path
        self._frame = frame
        self._params = {
            'include': include or [],
            'exclude': exclude or [],
        }
        self.frame = self._build_frame()

    def _build_frame(self):
        qs = urlencode(self._params, doseq=True)
        if qs:
            return f'{self.FILTER_FRAME}?{qs}'
        return self._frame

    def expand_path_with_frame(self, request, properties, path, frame):
        # Modified from snovault.util.expand_path.
        if isinstance(path, basestring):
             path = path.split('.')
        if not path:
            return
        name = path[0]
        remaining = path[1:]
        value = properties.get(name, None)
        if value is None:
            return
        if isinstance(value, list):
            for index, member in enumerate(value):
                if not isinstance(member, dict):
                    member = value[index] = request.embed(member, frame)
                self.expand_path_with_frame(request, member, remaining, frame)
        else:
            if not isinstance(value, dict):
                value = properties[name] = request.embed(value, frame)
            self.expand_path_with_frame(request, value, remaining, frame)

    def expand(self, request, properties):
        self.expand_path_with_frame(request, properties, self.path, self.frame)


def get_uuids_from_file(path):
    with open(path, 'r') as f:
        uuids = [
            uuid.strip()
            for uuid in f.readlines()
        ]
    return uuids



import pytest


@pytest.fixture(autouse=True)
def autouse_external_tx(external_tx):
    pass


@pytest.fixture
def lab(testapp):
    item = {
        'name': 'encode-lab',
        'title': 'ENCODE lab',
    }
    return testapp.post_json('/lab', item).json['@graph'][0]


@pytest.fixture
def remc_lab(testapp):
    item = {
        'name': 'remc-lab',
        'title': 'REMC lab',
    }
    return testapp.post_json('/lab', item).json['@graph'][0]


@pytest.fixture
def admin(testapp):
    item = {
        'first_name': 'Test',
        'last_name': 'Admin',
        'email': 'admin@example.org',
        'groups': ['admin'],
    }
    # User @@object view has keys omitted.
    res = testapp.post_json('/user', item)
    return testapp.get(res.location).json


@pytest.fixture
def wrangler(testapp):
    item = {
        # some objects have linkEnums
        'uuid': '4c23ec32-c7c8-4ac0-affb-04befcc881d4',
        'first_name': 'Wrangler',
        'last_name': 'Admin',
        'email': 'wrangler@example.org',
        'groups': ['admin'],
    }
    # User @@object view has keys omitted.
    res = testapp.post_json('/user', item)
    return testapp.get(res.location).json


@pytest.fixture
def submitter(testapp, lab, award):
    item = {
        'first_name': 'ENCODE',
        'last_name': 'Submitter',
        'email': 'encode_submitter@example.org',
        'submits_for': [lab['@id']],
        'viewing_groups': [award['viewing_group']],
    }
    # User @@object view has keys omitted.
    res = testapp.post_json('/user', item)
    return testapp.get(res.location).json


@pytest.fixture
def access_key(testapp, submitter):
    description = 'My programmatic key'
    item = {
        'user': submitter['@id'],
        'description': description,
    }
    res = testapp.post_json('/access_key', item)
    result = res.json['@graph'][0].copy()
    result['secret_access_key'] = res.json['secret_access_key']
    return result


@pytest.fixture
def viewing_group_member(testapp, award):
    item = {
        'first_name': 'Viewing',
        'last_name': 'Group',
        'email': 'viewing_group_member@example.org',
        'viewing_groups': [award['viewing_group']],
    }
    # User @@object view has keys omitted.
    res = testapp.post_json('/user', item)
    return testapp.get(res.location).json


@pytest.fixture
def remc_member(testapp, remc_lab):
    item = {
        'first_name': 'REMC',
        'last_name': 'Member',
        'email': 'remc_member@example.org',
        'submits_for': [remc_lab['@id']],
        'viewing_groups': ['REMC'],
    }
    # User @@object view has keys omitted.
    res = testapp.post_json('/user', item)
    return testapp.get(res.location).json


@pytest.fixture
def award(testapp):
    item = {
        'name': 'encode3-award',
        'rfa': 'ENCODE3',
        'project': 'ENCODE',
        'viewing_group': 'ENCODE',
    }
    return testapp.post_json('/award', item).json['@graph'][0]


@pytest.fixture
def invalid_award(testapp):
    item = {
        'name': 'award names cannot have spaces',
        'rfa': 'ENCODE3',
        'project': 'ENCODE',
        'viewing_group': 'ENCODE',
    }
    return testapp.post_json('/award' + '?validate=false', item).json['@graph'][0]


RED_DOT = """data:image/png;base64,iVBORw0KGgoAAAANSUhEUgAAAAUA
AAAFCAYAAACNbyblAAAAHElEQVQI12P4//8/w38GIAXDIBKE0DHxgljNBAAO
9TXL0Y4OHwAAAABJRU5ErkJggg=="""


@pytest.fixture
def attachment():
    return {'download': 'red-dot.png', 'href': RED_DOT}


@pytest.fixture
def snowball(testapp, lab, award):
    item = {
        'award': award['@id'],
        'lab': lab['@id'],
        'method': 'hand-packed',
    }
    return testapp.post_json('/snowball', item).json['@graph'][0]


@pytest.fixture
def invalid_snowball(testapp, lab, award):
    item = {
        'award': award['uuid'],
        'lab': lab['uuid'],
        'method': 'hand-packed',
        'schema_version': '1',
        'status': 'FLYBARGS',
    }
    return testapp.post_json('/snowball/' + '?validate=false', item).json['@graph'][0]


@pytest.fixture
def invalid_snowball2(testapp, lab, award):
    item = {
        'award': award['uuid'],
        'lab': lab['uuid'],
        'method': 'hand-packed',
        'schema_version': '1',
        'status': 'FLINGERNEERS',
        'alternate_accessions': ['THE THING THAT SHOULD NOT BE']
    }
    return testapp.post_json('/snowball/' + '?validate=false', item).json['@graph'][0]


@pytest.fixture
def snowflake(testapp, lab, award, snowball):
    item = {
        'snowset': snowball['@id'],
        'type': 'fluffy',
        'status': 'in progress',
        'award': award['@id'],
        'lab': lab['@id'],
    }
    return testapp.post_json('/snowflake', item).json['@graph'][0]


@pytest.fixture
def targets():
    targets = [
        {'name': 'one', 'uuid': '775795d3-4410-4114-836b-8eeecf1d0c2f'},
        {'name': 'two', 'uuid': 'd6784f5e-48a1-4b40-9b11-c8aefb6e1377'},
        {'name': 'quote:name', 'uuid': '0e627b3b-f5d2-41db-ac34-8e97bb8a028c'},
    ]
    return targets


@pytest.fixture
def sources():
    sources = [
        {
            'name': 'A',
            'target': 'one',
            'uuid': '16157204-8c8f-4672-a1a4-14f4b8021fcd',
            'status': 'current',
        },
        {
            'name': 'B',
            'target': 'two',
            'uuid': '1e152917-c5fd-4aec-b74f-b0533d0cc55c',
            'status': 'deleted',
        },
    ]
    return sources


@pytest.fixture
def posted_targets_and_sources(testapp, targets, sources):
    url = '/testing-link-targets/'
    for item in targets:
        testapp.post_json(url, item, status=201)

    url = '/testing-link-sources/'
    for item in sources:
        testapp.post_json(url, item, status=201)


@pytest.fixture
def posted_custom_embed_targets_and_sources(testapp, targets, sources):
    url = '/testing-custom-embed-targets/'
    for item in targets:
        testapp.post_json(url, item, status=201)

    url = '/testing-custom-embed-sources/'
    for item in sources:
        testapp.post_json(url, item, status=201)



'''
Simple uuid queue for indexing process
- Can be overridden by another script, class, module as long as it exposes
the proper functions used in the indexer classes.
'''
import time


class SimpleUuidServer(object):    #pylint: disable=too-many-instance-attributes
    '''Simple uuid queue as a list'''
    def __init__(self, queue_options):
        self._queue_options = queue_options
        self._start_us = int(time.time() * 1000000)
        self.queue_id = str(self._start_us)
        self._uuids = []
        self._uuid_count = 0
        self._worker_conns = {}
        self._worker_results = {}
        self._errors = []

    # Errors
    def add_errors(self, worker_id, errors):
        """Add errors after worker finishes"""
        errors_added = 0
        for batch_error in errors:
            batch_error['worker_id'] = worker_id
            errors_added += 1
            self._errors.append(batch_error)
        return errors_added

    def _has_errors(self):
        """Check if queue has errors"""
        if self._errors:
            return True
        return False

    def pop_errors(self):
        """Get and remove errors, they are expected to be handled after this"""
        errors = self._errors.copy()
        self._errors = []
        return errors

    # Worker
    @staticmethod
    def _get_blank_worker():
        return {
            'uuid_cnt': 0,
            'get_cnt': 0,
        }

    def _add_worker_conn(self, worker_id):
        """Add worker connection"""
        self._worker_results[worker_id] = []
        self._worker_conns[worker_id] = self._get_blank_worker()

    def get_worker_conns(self):
        """Return worker connection dicts"""
        return self._worker_conns

    def get_worker_conn_count(self):
        """Return number of worker conns"""
        return len(self._worker_conns)

    def _get_worker_id(self):
        time_id = self._start_us + self.get_worker_conn_count() + 1
        return str(time_id)

    def _get_worker_queue(self):
        """Queue to create workers with"""
        return self

    def get_worker(self):
        """Return the associated worker"""
        worker_id = self._get_worker_id()
        self._add_worker_conn(worker_id)
        worker_queue = self._get_worker_queue()
        return SimpleUuidWorker(
            self._queue_options,
            worker_id,
            worker_queue,
        )

    def update_worker_conn(self, worker_id, uuid_cnt, get_cnt):
        """Set worker conn info"""
        worker_conn = self._worker_conns[worker_id]
        worker_conn['uuid_cnt'] = uuid_cnt
        worker_conn['get_cnt'] = get_cnt

    def save_work_results(self, worker_id, results):
        """Save work results"""
        self._worker_results[worker_id].append(results)

    # Uuids
    def has_uuids(self, errs_cnt=0):  # pylint: disable=unused-argument
        '''Are there uuids in the queue'''
        if self._uuids:
            return True
        return False

    def _get_uuid(self):
        if self._uuids:
            return self._uuids.pop()
        return None

    def get_uuids(self, cnt, get_all=False):
        '''The only way to get uuids from queue'''
        uuids = []
        if cnt == -1 or get_all:
            cnt = len(self._uuids)
        if cnt:
            while cnt > 0:
                uuid = self._get_uuid()
                if not uuid:
                    break
                uuids.append(uuid)
                cnt -= 1
        return uuids

    def _load_uuid(self, uuid):
        '''
        Load uuid into queue
        - Uuid ordering impacts indexing time,
        list is reversed due to pop being used in
        get uuid
        '''
        self._uuids.insert(0, uuid)

    def load_uuids(self, uuids):
        '''
        The only way to load uuids in queue
        - Uuids are to be loaded once per indexing session
        '''
        if self.is_indexing():
            return None
        for uuid in uuids:
            self._load_uuid(uuid)
        return len(self._uuids)

    # Run
    def is_indexing(self, errs_cnt=0):  # pylint: disable=unused-argument
        '''Is an indexing process currently running'''
        if self.has_uuids(errs_cnt=errs_cnt) or self._has_errors():
            return True
        for worker_conn in self._worker_conns.values():
            if worker_conn['uuid_cnt']:
                return True
        return False

    def update_finished(self, given_worker_id, given_results):
        '''The way a worker finishes a batch of uuids'''
        msg = 'Worker id(%s) DNE.' % given_worker_id
        worker_conns = self.get_worker_conns()
        for worker_id, worker_conn in worker_conns.items():
            if worker_id == given_worker_id:
                errors_added = self.add_errors(worker_id, given_results['errors'])
                uuid_cnt = worker_conn['uuid_cnt'] - given_results['successes'] - errors_added
                self.update_worker_conn(
                    worker_id,
                    uuid_cnt,
                    worker_conn['get_cnt'],
                )
                self.save_work_results(worker_id, given_results)
                if uuid_cnt != 0:
                    msg = 'Queue Server cannot close worker (%s)' % worker_id
                else:
                    msg = 'Okay'
                break
        return msg

    def close_indexing(self):
        '''Close indexing sessions'''
        pass


class SimpleUuidWorker(object):  #pylint: disable=too-many-instance-attributes
    '''Basic uuid worker to get uuids for indexing'''
    def __init__(self, queue_options, worker_id, queue):
        self.queue_options = queue_options
        self.worker_id = worker_id
        self._queue = queue
        self.is_running = False
        # Worker conn info
        self.get_cnt = 0
        self.uuid_cnt = 0
        # MPIndexer
        self.processes = self.queue_options['processes']
        self.chunk_size = self.queue_options['chunk_size']

    # Uuids
    def get_uuids(self, get_all=False):
        '''Get all or some of uuids'''
        self.get_cnt += 1
        uuids = []
        if self.uuid_cnt == 0:
            uuids = self._queue.get_uuids(self.queue_options['batch_size'], get_all=get_all)
            self.uuid_cnt = len(uuids)
        self._queue.update_worker_conn(
            self.worker_id,
            self.uuid_cnt,
            self.get_cnt,
        )
        return uuids

    # Run
    def update_finished(self, results):
        '''Update server with batch results'''
        msg = self._queue.update_finished(self.worker_id, results)
        if msg == 'Okay':
            self.uuid_cnt = 0
            return None
        return 'Update finished could not reset worker: {}'.format(msg)



'''Local storage redis Fixture'''
import os
import sys

import subprocess


# Commented out config is added in 'initdb' below
REDIS_DEFAULTS = [
    # 'bind 127.0.0.1 ::1',                         Disabled ipv6 since circleci does not support
    'protected-mode yes',
    # port 6379',                                   Port change
    'tcp-backlog 511',
    'timeout 0',
    'tcp-keepalive 300',
    'daemonize no',
    'supervised no',
    # 'pidfile /var/run/redis_6379.pid',            Port change
    'loglevel notice',
    'logfile ""',
    'databases 16',
    'always-show-logo yes',
    'save 900 1',
    'save 300 10',
    'save 60 10000',
    'stop-writes-on-bgsave-error yes',
    'rdbcompression yes',
    'rdbchecksum yes',
    'dbfilename dump.rdb',
    # 'dir /usr/local/var/db/redis/',               Variable database location
    'slave-serve-stale-data yes',
    'slave-read-only yes',
    'repl-diskless-sync no',
    'repl-diskless-sync-delay 5',
    'repl-disable-tcp-nodelay no',
    'slave-priority 100',
    'lazyfree-lazy-eviction no',
    'lazyfree-lazy-expire no',
    'lazyfree-lazy-server-del no',
    'slave-lazy-flush no',
    'appendonly no',
    'appendfilename "appendonly.aof"',
    'appendfsync everysec',
    'no-appendfsync-on-rewrite no',
    'auto-aof-rewrite-percentage 100',
    'auto-aof-rewrite-min-size 64mb',
    'aof-load-truncated yes',
    'aof-use-rdb-preamble no',
    'lua-time-limit 5000',
    'slowlog-max-len 128',
    'latency-monitor-threshold 0',
    'notify-keyspace-events ""',
    'hash-max-ziplist-entries 512',
    'hash-max-ziplist-value 64',
    'list-max-ziplist-size -2',
    'list-compress-depth 0',
    'set-max-intset-entries 512',
    'zset-max-ziplist-entries 128',
    'zset-max-ziplist-value 64',
    'hll-sparse-max-bytes 3000',
    'activerehashing yes',
    'client-output-buffer-limit normal 0 0 0',
    'client-output-buffer-limit slave 256mb 64mb 60',
    'client-output-buffer-limit pubsub 32mb 8mb 60',
    'hz 10',
    'aof-rewrite-incremental-fsync yes',
]


def initdb(datadir, redis_port, echo=False):
    '''Create redis db config in data dir'''
    redis_config_lines = REDIS_DEFAULTS.copy()
    redis_config_lines.append(f"port {redis_port}")
    redis_config_lines.append(f"pidfile /var/run/redis_{redis_port}.pid")
    redis_config_lines.append(f"dir {datadir}")
    redis_config_lines.append('bind 127.0.0.1')
    redis_config_path = f"{datadir}/redis.conf"
    if not os.path.exists(datadir):
        os.makedirs(datadir)
    with open(redis_config_path, '+w') as file_handler:
        file_handler.writelines('\n'.join(redis_config_lines))
    if echo:
        print(f"Redis Config created: {redis_config_path}")
    return redis_config_path


def server_process(redis_config_path, redis_port, redis_index, echo=False):
    '''Start redis server'''
    args = [
        os.path.join('redis-server'),
        f"{redis_config_path}",
    ]
    if echo:
        print(f"Starting redis server: {' '.join(args)}")
        print(f"Connect with 'redis-cli -p {redis_port} -n {redis_index}'")
    redis_process = subprocess.Popen(
        args,
        close_fds=True,
        stdout=subprocess.PIPE,
        stderr=subprocess.STDOUT,
    )
    return redis_process


def _main():
    import atexit
    import shutil
    import tempfile
    datadir = tempfile.mkdtemp()

    def _cleanup():
        shutil.rmtree(datadir)
        print(f"Redis Cleaned dir: {datadir}")

    try:
        print(f"Starting in dir: {datadir}")
        redis_config_path = initdb(datadir, 6378, echo=True)
        redis_process = server_process(redis_config_path, 6378, echo=True)
    except Exception as ecp:  # pylint: disable=broad-except
        _cleanup()
        shutil.rmtree(datadir)
        print('Cleaned dir: %s' % datadir)
        raise ecp

    @atexit.register
    def cleanup_process():  # pylint: disable=unused-variable
        '''System Exit hook to remove db and kill redis server process'''
        try:
            if redis_process.poll() is None:
                redis_process.terminate()
                for line in redis_process.stdout:
                    sys.stdout.write(line.decode('utf-8'))
                redis_process.wait()
        finally:
            _cleanup()

    try:
        break_line = b'Ready to accept connections'
        for line in iter(redis_process.stdout.readline, b''):
            sys.stdout.write(line.decode('utf-8'))
            strip_line = line.strip()
            if strip_line.endswith(break_line) and break_line:
                print('Redis Started. ^C to exit.')
                break_line = None
    except KeyboardInterrupt:
        raise SystemExit(0)


if __name__ == '__main__':
    _main()
